% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(random_permutations, [get_num_bits/3,
                                get_masks/3,
                                random_permutation_element/10,
                                init_random_permutations/0
                                ]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,cbc).
:- module_info(description,'This module provides on-the-fly construction of random permutations of intervals.').

foreign_resource(random_permutations,[get_num_bits,draw_index]).
foreign(get_num_bits, c, get_num_bits(+integer,-integer,-integer)).
foreign(draw_index, c, draw_index(+integer,+integer,+integer,+integer,+integer,+integer,+integer,+integer,-integer,-integer)).

:- dynamic is_initialised/0.

init_random_permutations :- is_initialised,!.
init_random_permutations :-
    load_foreign_resource(library(random_permutations)),
    assert(is_initialised).

get_masks(HalfNumBits,LeftMask,RightMask) :-
    RightMask is (1 << HalfNumBits) - 1,
    LeftMask is RightMask << HalfNumBits.

random_permutation_element(Index,MaxIndex,From,To,Seed,NumBits,LeftMask,RightMask,RandomElement,NextIndex) :-
    draw_index(Index,MaxIndex,Seed,NumBits,LeftMask,RightMask,From,To,DrawnElement,NextIndex),
    % working on a 4^x long interval. thus, we might pick a number that is too large
    % if this happens, we just pick a new one
    % to avoid context switching overhead, this is now done inside the C code
    RandomElement is DrawnElement + From.

% -------------------
% Testing predicates
% -------------------
% unused at the moment:
%:- public shuffle/2.
%:- use_module(library(system)).
%shuffle(From,To) :-
%    init_random_permutations,
%    IntervalLength is To - From + 1,
%    get_num_bits(IntervalLength,MaxIndex,NumBits),
%    get_masks(NumBits,LeftMask,RightMask),
%    now(Seed),
%    shuffle(From,0,To,MaxIndex,Seed,NumBits,LeftMask,RightMask).
%shuffle(From,CurIdx,To,MaxIndex,Seed,NumBits,LeftMask,RightMask) :-
%    random_permutation_element(CurIdx,MaxIndex,From,To,Seed,NumBits,LeftMask,RightMask,Drawn,NextIdx),
%    format('Drawing index ~w resulted in ~w~n',[CurIdx,Drawn]),
%    shuffle(From,NextIdx,To,MaxIndex,Seed,NumBits,LeftMask,RightMask).
%shuffle(_,_,_,_,_,_,_,_).
