import org.zeromq.ZMQ;
import org.zeromq.ZMsg;
import org.zeromq.ZFrame;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.StringBuffer;
import java.util.List;
import java.util.ArrayList;
// Compile: javac -cp /path/to/jzmq/src/ StatLeecher.java
// Execute: java -ea -Djava.library.path=/path/to/libjzmq -cp /path/to/jzmq/src/:. StatLeecher



public class ZMQMaster {
    private static ZMQ.Socket hash_publish,
    hash_collect, 
    work_publish, 
    work_collect,
    id_response,
    send_ctrl,
    recv_result;

    private static BufferedWriter stats = null;
    private static List<String> statistics = new ArrayList<String>();

    private static boolean running = true;
    private final static int TIMEINDEX = 8;

    private static ZMQ.Context context;
    private static ZMQ.Poller p;

    private final static int N_WORKERS = 16;
    private final static int FIRST_WORKER = 2;

    private final static int minimum_size = 200;

    private static int next_worker = 2;
    private volatile static int[] queues = new int[N_WORKERS];
    private static int numberOfWP = 0;
    private static String model;

    public static void main(String[] args) {
        model = "DA" + args[0];
        init();
        final Balancer b = new Balancer();
        b.start();

        try {
            stats.write("[ ");
        } catch (final IOException e) {
            System.err.println("Could not write to file");
        }

        while (running) {
            p.poll(1);

            for (int i = 0; i < 4; i++) {
                final boolean newMessage = p.pollin(i);
                if (newMessage) {
                    switch(i) {
                        case 0:
                            running = h_hash(p.getSocket(i));
                            break;
                        case 1: h_work  (p.getSocket(i)); break;
                        case 2: h_id    (p.getSocket(i)); break;
                        case 3: h_result(p.getSocket(i)); break;
                    }
                }
            }       
        }
        b.setRunning(false);
        ZMsg.newStringMsg("TERM").send(send_ctrl);
        cleanup();

    }

    private static void init() {
        try {
            final FileWriter fstream = new FileWriter("statistics", true);
            stats = new BufferedWriter(fstream);
        } catch (final IOException e) {
            System.err.println("Could not open file.");
        }

        for (int i = 0; i < N_WORKERS; i++)
            queues[i] = -1;
        
        context = ZMQ.context(1);

        hash_publish = context.socket(ZMQ.PUB);
        hash_collect = context.socket(ZMQ.PULL);
        work_publish = context.socket(ZMQ.ROUTER);
        work_collect = context.socket(ZMQ.ROUTER);
        id_response  = context.socket(ZMQ.REP);
        send_ctrl    = context.socket(ZMQ.PUB);
        recv_result  = context.socket(ZMQ.PULL);

        hash_publish.bind("tcp://*:5000");
        hash_collect.bind("tcp://*:5001");
        work_publish.bind("tcp://*:5002");
        work_collect.bind("tcp://*:5003");
        id_response.bind ("tcp://*:5005");
        send_ctrl.bind   ("tcp://*:5007");
        recv_result.bind ("tcp://*:5008");

        p = context.poller(4);
        p.register(hash_collect, 1);
        p.register(work_collect, 1);
        p.register(id_response,  1);
        p.register(recv_result,  1);
    }

    public static boolean checkShutdown() {
        if (numberOfWP == 0)
            return false;
        
        boolean shutdown = true;
        boolean started = false;
        int i;
        for (i = 0; i < N_WORKERS; i++) {
            if (queues[i] == 0) {
                started = true;
            }
            if (queues[i] > 0) {
                shutdown = false;
                break;
            }
        }
        return shutdown && started;
    }

    public static boolean h_hash(final ZMQ.Socket socket) {
        final ZMsg msg = ZMsg.recvMsg(socket, 0);
        final int size = msg.size();
        final ZMsg out = new ZMsg();
        int send = put_hashes(msg, out);

        final String id_q = msg.popString();
        final String[] split = id_q.split(" ");
        final int worker = Integer.parseInt(split[0]);
        final int qs     = Integer.parseInt(split[1]);
        
        queues[worker] = qs;
        final String runs = msg.popString();
        
        // TODO: write more to file
        try {
            writeTimeStatistics(msg, worker, runs, qs);
        } catch (final IOException e) {
            System.err.println("Error writing to file");
        }

        if (send > 0) {
            out.send(hash_publish);
        }
        msg.destroy();
        out.destroy();

        if (checkShutdown())
            return false;

        if (queues[FIRST_WORKER] == -2) {
            pok();
        }
    

        return true;
    }

    private static int put_hashes(final ZMsg in, final ZMsg out) {
        final int size = in.size();
 
        assert(size >= TIMEINDEX + 5); // may be 3?
        int i = 0;
        for (; i < size - TIMEINDEX - 5; i++) {
            final ZFrame frame = in.pop();
            assert(frame.size() == 21);
            final byte[] digest = frame.getData();
            out.add(digest);
            // TODO: put digest in hashmap
            // hashes.put(digest);
            frame.destroy();
        }
        // TODO: maximum number of states to check
        return i;
    }

    private static void writeTimeStatistics(final ZMsg msg, final int worker, final String runs, final int qs) throws IOException {
        //stats.write(worker + "; " + runs + "; " qs + "; "); 
        assert(msg.size() == TIMEINDEX + 3);
        final StringBuffer sb = new StringBuffer();
        final String out = "{ :id " + worker + " :runs " + runs + " :qs " + qs + " ";
        sb.append(out);
        final String[] labels = { ":prolog", ":workhard", ":emptyq", ":overall", ":hash", ":req", ":work", ":succ" };
        for (int i = 0; i < TIMEINDEX; i++) {
            final String pl = msg.popString();
            final int pos = pl.indexOf('.');
            final String sub = pl.substring(0, pos + 4);
            sb.append(labels[i] + " " + sub + " ");
        }
        final String timestamp = " :timestamp " + msg.popString();
        final String work_hard_runs = " :whruns " + msg.popString();
        final String digest = " :digest " + msg.popString();
        sb.append(timestamp + work_hard_runs + digest + " }");
        final String statLine = sb.toString();
        statistics.add(statLine);
        stats.write(statLine);
        stats.newLine();
    }

    private static void pok() {
        final ZMQ.Socket back = context.socket(ZMQ.DEALER);
        back.connect("tcp://localhost:5003");
        final String root = "DAroot";
        final byte[] rootBytes = root.getBytes();
        final byte[] digest = toSHA1(root);
        final byte[] wp = new byte[digest.length + rootBytes.length];

        for (int i = 0; i < digest.length; i++) {
            wp[i] = digest[i];
        }
        for (int i = digest.length; i < wp.length; i++) {
            wp[i] = rootBytes[i-digest.length];
        }
     
        final ZMsg msg = new ZMsg();
        msg.add(String.valueOf(FIRST_WORKER));
        msg.add(wp);
     
        msg.send(back);
        back.disconnect("tcp://localhost:5003");
        back.close();

        msg.destroy();
    }

    public static byte[] toSHA1(String convertme) {
        MessageDigest md = null;
        try {
            md = MessageDigest.getInstance("SHA-1");
        }
        catch(NoSuchAlgorithmException e) {
            e.printStackTrace();
        } 

        return md.digest(convertme.getBytes());
    }


    public static void h_work(final ZMQ.Socket socket) {
        numberOfWP++;
        final ZMsg msg = ZMsg.recvMsg(socket);
        msg.pop().destroy(); // frame containing worker id "from"
        // TODO: distribution via master
        msg.send(work_publish);
    }

    public static void h_id(final ZMQ.Socket socket) {
        socket.recv(0);
        final String id = String.valueOf(next_worker);
        next_worker++;
        final ZMsg msg = new ZMsg();
        msg.addString(id);
        msg.addString(model);
        msg.send(socket);
        msg.destroy();
    }

    public static void h_result(final ZMQ.Socket socket) {
        final ZMsg msg = ZMsg.recvMsg(socket);
        while (0 < msg.size()) {
            System.out.println(msg.popString());
       }
    }

    private static void cleanup() {
        hash_publish.close();
        hash_collect.close(); 
        work_publish.close();
        work_collect.close();
        id_response.close();
        send_ctrl.close();
        recv_result.close();
        context.term();

        try {
            stats.write("]");
            stats.newLine();
            stats.write((next_worker - 2) + " worker(s)");
            stats.newLine();
            stats.write("----------------------------");
            stats.newLine();
            stats.flush();
            stats.close();
        } catch (final IOException e) {
            System.err.println("Error closing filestream");
        }
    }




    private static class Balancer extends Thread {
        private volatile boolean running = true;
        public void run() {
            int p = 0;
            while (running) {
                p++;
                balanceWork();
                if (p % 5 == 0) {
                    printQueues();
                }
                try {
                    Thread.sleep(200);
                } catch (final InterruptedException e) {
                    setRunning(false);
                }
            }
        }

        public void setRunning(boolean b) {
            running = b;
        }

        private void balanceWork() {
            if (next_worker - FIRST_WORKER > 1) {
                int need = 0, has = 0;
                for (int i = FIRST_WORKER; i < next_worker; i++) {
                    if (queues[i] > queues[has])
                        has = i;
                    if (queues[i] == 0 || queues[i] == -2)
                        need = i;
                }

                if (queues[has] > minimum_size && need > 0) {
                    final ZMsg msg = new ZMsg();
                    msg.addString(String.valueOf(has));
                    msg.addString(String.valueOf(need));
                    msg.addString(String.valueOf(queues[has]/2));
                    msg.send(work_collect);
                    msg.destroy();
                }
            }
        }

        private void printQueues() {
            for (int i = FIRST_WORKER; i < next_worker; i++) {
                System.out.print("(" + i + ": " + queues[i] + ") ");
            }
            System.out.println();
        }
    }
}
