#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "../../../lib/include/zmq.h"
#include "../../../lib/include/czmq.h"

#include "../../../common/ring.h"

#include "mhash.h"
#include "mid.h"
#include "mctrl.h"
#include "mbalancer.h"
#include "master.h"
#include "mprint.h"
#include "mbeat.h"

#include "../../../common/hashrecovery.h"

static void init_zmq(zctx_t **ctx, int maxstate, int len,
                     int port_start, int debug_level, char *logfile,
                     int recover, int strict_mode);
static void teardown_zmq(zctx_t *ctx);

static char *fwp = NULL;
static int fwp_len;

extern void *global_printer;

volatile int running = 1;
volatile int nmbr_workers;
int assertion_mode = -1;

int master_main(int len, int maxstate, int port_start,
                int assertion, int recover, int debug_lvl,
                int strict_mode, char *logfile) {
    printf("recovery mode: %d\n", recover);
    nmbr_workers = recover;
    debug_level = debug_lvl;

    zctx_t *ctx;
    init_zmq(&ctx, maxstate, len, port_start, debug_level, logfile, recover, strict_mode);
    // running a reactor
    teardown_zmq(ctx);

    return 0;
}

static void init_zmq(zctx_t **ctx, int maxstate, int len,
                     int port_start, int debug_level, char *logfile,
                     int recover, int strict_mode) {
    *ctx = zctx_new ();
    zmq_pollitem_t worker_printer;
    init_master_printer(*ctx, debug_level, &worker_printer, port_start, logfile);

    // no poller
    init_send_ctrl(*ctx, port_start);

    zmq_pollitem_t join_requests;
    init_id_rep(*ctx, &join_requests, port_start);

    zmq_pollitem_t hash_poller;
    char *filename = init_hash_module(*ctx, &hash_poller, maxstate, port_start);

    //init_mbeat(*ctx);

    zmq_pollitem_t stats_poller;
    init_work_balancer(*ctx, &stats_poller, len, port_start);

    if (assertion_mode == -1) {
        assert(fwp);
        // regular model checking
        zstr_send(global_printer, "{:type :info :mode :model}\n");
        if (recover) {
#ifndef TMPFOLDER
#define TMPFOLDER "/var/tmp/"
#endif
            recover_hashes(TMPFOLDER, *ctx, filename);
            int empty_workers = receive_hashes(recover, *ctx, port_start);
            set_empty_workers(empty_workers);
        } else {
            start_work(fwp, fwp_len);
        }
    } else {
        // assertion checking
        start_work_assertion(assertion_mode);
    }


    pthread_t hashthread;
    start_h_mhash(&hashthread, &strict_mode);

    pthread_t idthread;
    start_h_joiner(&idthread);

    pthread_t balancer;
    start_work_balancer(&balancer);

    pthread_t printer;
    start_printer(&printer);

    void *retval;
    pthread_join(hashthread, &retval);

    running = 0;

    pthread_detach(printer);
    pthread_detach(balancer);
    pthread_detach(idthread);

    pthread_cancel(printer);
    pthread_cancel(balancer);
    pthread_cancel(idthread);
    
    send_term();
}

void set_fwp(char *term, int len) {
    fwp = (char*) malloc(len);
    assert(fwp);
    memcpy(fwp, term, len);
    fwp_len = len;
}

static void teardown_zmq(zctx_t *ctx) {
    teardown_id_rep(ctx);

    //teardown_mbeat(ctx);
    teardown_hash_module(ctx);
    teardown_work_balancer(ctx);
    teardown_send_ctrl(ctx);
    stop_master_printer(ctx);
}


void set_assertion_mode(int number_of_assertions) {
    assertion_mode = number_of_assertions;
}
