#include <stdarg.h>
#include <pthread.h>

#include "../../../lib/include/zmq.h"
#include "../../../lib/include/czmq.h"

#include <stdio.h>
#include "mprint.h"

static void *psocket = NULL;
static void *socket_out = NULL;
static void *worker_socket = NULL;
static pthread_t printthread;
static FILE *logfile = NULL;
void *global_printer = NULL;

int debug_level = 0;
extern volatile int running;

static void *loop(void *arg) {
    while(running) {
        if (zsocket_poll(psocket, 200)) {
            zmsg_t *msg = zmsg_recv(psocket);
            
            char *str;
            while ((str = zmsg_popstr(msg)) != NULL) {
                fprintf(logfile, "%s", str);
                fflush(logfile);
                free(str);
            }
            zmsg_destroy(&msg);
        }
    }

    while (zsocket_poll(psocket, 1)) {
        zmsg_t *msg = zmsg_recv(psocket);
        
        char *str;
        while ((str = zmsg_popstr(msg)) != NULL) {
            fprintf(logfile, "%s", str);
            fflush(logfile);
            free(str);
        }
        zmsg_destroy(&msg);
    }

    return NULL;
}

void init_master_printer(zctx_t *ctx, int debug_lvl, zmq_pollitem_t *poller,
                  int port_start, char *target) {
    logfile = fopen(target, "a+");
    psocket = zsocket_new(ctx, ZMQ_PULL);
    zsocket_bind(psocket, "inproc://printer");
    socket_out = zsocket_new(ctx, ZMQ_PUSH);
    global_printer = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(socket_out, "inproc://printer");
    zsocket_connect(global_printer, "inproc://printer");

    worker_socket = zsocket_new(ctx, ZMQ_PULL);
    zsocket_bind(worker_socket, "tcp://*:%d", port_start + 6);
    *poller = (zmq_pollitem_t) { worker_socket, 0, ZMQ_POLLIN };

    if (debug_lvl) {
        pthread_create(&printthread, NULL, loop, NULL);
        debug_level = debug_lvl;
    }

}

void stop_master_printer(zctx_t *ctx) {
    while (zsocket_poll(worker_socket, 100)) {
        zmsg_t *msg = zmsg_recv(worker_socket);
        zmsg_send(&msg, socket_out);
        zmsg_destroy(&msg);
    }
    if (debug_level) {
        pthread_join(printthread, NULL);
    }
    fflush(logfile);
    fclose(logfile);
    logfile = NULL;
    zsocket_destroy(ctx, psocket);
    zsocket_destroy(ctx, socket_out);
    zsocket_destroy(ctx, global_printer);
    zsocket_destroy(ctx, worker_socket);
}


static void *h_prints (void *arg) {
    while (running) {
        zmsg_t *msg = zmsg_recv(worker_socket);
        zmsg_send(&msg, socket_out);
        zmsg_destroy(&msg);
    }
    pthread_exit(NULL);
}


void start_printer(pthread_t *thread) {
    pthread_create(thread, NULL, h_prints, NULL);
}
