#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <math.h>
#include "CuTest.h"
#include "../common/hashtrie.h"
#include "../common/sha.h"

void TestTrie(CuTest *tc) {
    init_hashtrie();

    trie_put_checked("012345678901234567890"); 
    CuAssertTrue(tc, trie_contains("012345678901234567890")); 
    CuAssertTrue(tc, trie_contains_processed("012345678901234567890")); 
    trie_put_unchecked("012345678901234567890"); 
    CuAssertTrue(tc, trie_contains("012345678901234567890")); 
    trie_put_unchecked("012345678201234567890"); // first '9' is a '2'
    CuAssertTrue(tc, trie_contains("012345678901234567890")); 
    CuAssertTrue(tc, trie_contains("012345678201234567890")); 
    CuAssertTrue(tc, trie_contains_processed("012345678901234567890")); 

    CuAssertIntEquals(tc, 1, trie_get_checked_states_count());
    CuAssertIntEquals(tc, 2, trie_count_elements());
    int a;
    for (a = 1; a <= 50000; a++) {        
        int len = (int)log10(a) + 2;
        char *res = (char*) malloc(len);
        sprintf(res,"%i",a);
        char *digest = (char*) malloc(20);
        sha1(res, len, digest);
        trie_put_unchecked(digest);
        free(digest);
        free(res);
    }
    CuAssertIntEquals(tc, 1, trie_get_checked_states_count());
    CuAssertIntEquals(tc, 50002, trie_count_elements());

/*    char *hashes = trie_iterate();
    for (a = 0; a < all_states; a++) {
        int res = trie_contains(hashes + (a * 21) + 1);
        assert(hashes[a * 21] == 0 || hashes[a * 21] == 1);
        if (hashes[a * 21] == 0)
            assert(res == -1);
        else
            assert(res == 1);
    }

    free(hashes);*/

//    printf("# of elements: %d\n", teardown_and_dump());
    CuAssertIntEquals(tc, 50002, teardown_hashtrie());
}

CuSuite *TrieGetSuite() {
    CuSuite *suite = CuSuiteNew();
    SUITE_ADD_TEST(suite, TestTrie);
    return suite;
}
