#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "../../../lib/include/zmq.h"
#include "../../../lib/include/czmq.h"

#include "../../../common/hashtrie.h" // for print_readable_digest
#include "../../../common/queue.h"
#include "../../../common/shm_common.h"

#include "whash_send.h"
#include "wdebug.h"



static void *send_hashes = NULL;
// this message contains results and successors
static zmsg_t *msg = NULL;

void init_send_hashes(zctx_t *ctx, char *masteraddr, int port) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :init_send_hashes");
    assert(send_hashes == NULL);
    send_hashes = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(send_hashes,"ipc:///tmp/zmq-proxy-%i", port+1);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :init_send_hashes");
}

void teardown_send_hashes(zctx_t *ctx) {
    assert(send_hashes != NULL);
    zsocket_destroy(ctx, send_hashes);
    send_hashes = NULL;
}


void init_successors(void) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :init_successors");
    assert(msg == NULL);
    msg = zmsg_new();
    zmsg_addstr(msg, "Hashes:");
    assert(msg != NULL);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :init_successors");
}


static void add_digest_processed(char *digest) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,
                char d[TRIE_HASHLEN * 2 + 1]; print_readable_digest((unsigned char*) digest+1,d);,,
                ":enter :add_digest_processed :digest :%s", d);
    assert(msg != NULL);
    char pl[TRIE_DATALEN];
    memcpy (pl, digest, TRIE_HASHLEN);
    pl[TRIE_HASHLEN] = (unsigned char) 128;
    memset(pl + TRIE_HASHLEN + 1, 0, TRIE_ADDITIONAL_DATA - 1);
    zmsg_addmem(msg, pl, TRIE_DATALEN);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :add_digest_processed");
}

void send_successors(tCell *cell) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,
                char d[TRIE_HASHLEN * 2 + 1]; print_readable_digest((unsigned char*) (cell->digest)+1,d);,,
                ":enter :send_successors :digest :%s", d);
    assert(msg != NULL);
    add_digest_processed(cell->digest);
    zmsg_pushstr(msg, "Results:");
    zmsg_send(&msg, send_hashes);
    zmsg_destroy(&msg);

    msg = NULL;
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :send_successors");
}

void add_queued_digest(char string[TRIE_HASHLEN], unsigned char *infos) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,
                char d[TRIE_HASHLEN * 2 + 1]; print_readable_digest((unsigned char*) string,d);,,
                ":enter :add_queued_digest :digest :%s", d);
    assert(msg != NULL);
    char tmp[TRIE_DATALEN];
    memcpy (tmp, string, TRIE_HASHLEN);
    memcpy (tmp + TRIE_HASHLEN, infos, TRIE_ADDITIONAL_DATA);
    zmsg_addmem(msg, tmp, TRIE_DATALEN);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :add_queued_digest");
}

void put_result(char *term) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :put_result");
    zmsg_pushstr(msg, term);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :put_result");
}
