% b_arithmetic_expressions.pl
% returns for arithmetic expressions a symbolic value that can be called in CLP(FD)

b_compute_arith_expression(b(Expr,Type,Info),LS,S,R,WF) :-  !,
    ((ground(Type), Type  == integer) -> true
     ; add_error(b_compute_arith_expression,'Arithmetic expression has illegal type: ',b(Expr,Type,Info))),
     b_compute_arith_expression2(Expr,Info,LS,S,R,WF).

b_compute_arith_expression(Expr,LS,S,R,WF) :-
  add_error(b_compute_arith_expression,'Expression not properly wrapped: ',Expr),
  b_compute_arith_expression2(Expr,unknown,LS,S,R,WF).


b_compute_arith_expression2(unary_minus(Arg1),_I,LocalState,State,Value,WF) :- !,
        b_compute_arith_expression(Arg1,LocalState,State,SV1,WF),
        Value = '-'(SV1).
b_compute_arith_expression2(add(Arg1,Arg2),_I,LocalState,State,Value,WF) :- !,
        b_compute_arith_expression(Arg1,LocalState,State,SV1,WF),
        b_compute_arith_expression(Arg2,LocalState,State,SV2,WF),
        Value = '+'(SV1,SV2).
b_compute_arith_expression2(minus(Arg1,Arg2),_I,LocalState,State,Value,WF) :- !,
        b_compute_arith_expression(Arg1,LocalState,State,SV1,WF),
        b_compute_arith_expression(Arg2,LocalState,State,SV2,WF),
        Value = '-'(SV1,SV2).
b_compute_arith_expression2(multiplication(Arg1,Arg2),_I,LocalState,State,Value,WF) :- !,
        b_compute_arith_expression(Arg1,LocalState,State,SV1,WF),
        b_compute_arith_expression(Arg2,LocalState,State,SV2,WF),
        Value = '*'(SV1,SV2).
b_compute_arith_expression2(Expr,Info,LocalState,State,Value,WF) :-
        b_compute_expression2(Expr,integer,Info,LocalState,State,int(Value),WF).
        %b_compute_expression2(Expr,integer,Info,LocalState,State,R,WF), R=int(Value). % this version does not seem faster
      %  kernel_objects:basic_type2(integer,int(Value)).

% be sure to register all cases here in clpfd_arith_integer_expression in b_ast_cleanup


    % Note: calling b_compute_arith_expression will also instantiate a variable to at least the int(_) skeleton; thereby enabling propagation
    %       this is potentially better than calling the default versions of the predicates, which may wait until the int(_) skeleton is set up before propagation


:- use_module(clpfd_interface,[clpfd_eq_expr_optimized/2]).

b_test_arith_equal_boolean_expression(Arg1,Arg2,LocalState,State,WF) :-
    b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
    b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
    clpfd_eq_expr_optimized(CLPFD_Expr1,CLPFD_Expr2).

%(preferences:preference(use_smt_mode,true)-> clpfd_eq_expr_optimized(CLPFD_Expr1,CLPFD_Expr2)
%      ; clpfd_eq_expr(CLPFD_Expr1,CLPFD_Expr2)).
% maybe we should call clpfd_eq_expr_optimized only in SMT mode and call clpfd_eq_expr otherwise ??
% initially test 1077 ran considerably slower with clpfd_eq_expr_optimized
