% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% WARNING (DEPRECATED) : it is planned to replace the debug_properties commands using the unsat_cores.pl predicates
% operation enabling code should be moved to a separate module

:- module(predicate_debugger,[tcltk_debug_properties/3,
  tcltk_debug_properties_or_op/4,  adapt_tcl_operation_name/2,
  get_operation_propositional_guards/3, get_operation_propositional_guards/6,
  get_operation_enabling_condition/7,
  get_unsimplified_operation_guard/2,
  get_unsimplified_operation_enabling_condition/5,
  get_simplified_operation_enabling_condition/5,
  tcltk_get_unsat_core_with_types/1]).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,misc).
:- module_info(description,'The predicate debugger adds conjuncts one by one until the predicate becomes unsatisfiable.').

:- use_module(debug).
:- use_module(self_check).
:- use_module(preferences).
:- use_module(error_manager).

:- use_module(translate).
:- use_module(b_interpreter).
:- use_module(bsyntaxtree).
:- use_module(b_ast_cleanup).

:- use_module(library(timeout)).
:- use_module(library(lists)).


:- meta_predicate maplist5(4,-,-,-,-).
maplist5(_P,[],[],[],[]).
maplist5(P,[H1|T1],[H2|T2],[H3|T3],[H4|T4]) :-
     if(call(P,H1,H2,H3,H4),
        maplist5(P,T1,T2,T3,T4),
        (add_internal_error('Call fails: ',maplist5(P,H1,H2,H3,H4)),
         fail)).

/* b_interpreter:b_debug_properties. */

:- volatile checked_property/1, last_checked_solution/2, failed_property/2, time_out_property/2, cur_conjunct_nr/1, projection_onto_component/1.
:- dynamic checked_property/1.
:- dynamic last_checked_solution/2.
:- dynamic failed_property/2.
:- dynamic time_out_property/2.
:- dynamic cur_conjunct_nr/1.
:- dynamic projection_onto_component/1.

reset_debug_properties :-
   retractall(last_debug_info(_,_)),
   retractall(checked_property(_)),
   retractall(last_checked_solution(_,_)),
   retractall(failed_property(_,_)),
   retractall(time_out_property(_,_)),
   retractall(cur_conjunct_nr(_)),
   retractall(projection_onto_component(_)).

:- use_module(b_global_sets,[b_fd_type/3]).
tcltk_debug_properties(RealRes,UnsatCore,Satisfiable) :-
   tcltk_debug_properties_or_op('@PROPERTIES',UnsatCore,RealRes,Satisfiable).

% code below because names with $ cause problems in Tcl with eval
adapt_tcl_operation_name(X,R) :- var(X),!,R=X.
adapt_tcl_operation_name('@PROPERTIES','$setup_constants') :- !.
adapt_tcl_operation_name('@INITIALISATION','$initialise_machine') :- !.
adapt_tcl_operation_name(X,X).

tcltk_debug_properties_or_op(TclOpName,ComputeUnsatCore,list(RealRes),Satisfiable) :-
   adapt_tcl_operation_name(TclOpName,OpName),
   print(tcltk_debug_properties_or_op(OpName,ComputeUnsatCore,RealRes)),nl,
   (ComputeUnsatCore==false ->
      (OpName='$setup_constants' -> b_debug_properties
       ; OpName = '@GOAL' -> b_debug_goal
       ; b_debug_operation(OpName))
    ;    set_fast_unsat_core(ComputeUnsatCore),
         print('*** Finding Unsat Core : '), print(ComputeUnsatCore),nl,
         (b_find_unsat_core
          -> print('*** Finished '),nl
          ; add_error_fail(predicate_debugger,'Finding Unsat Core failed',OpName))
   ),
   print(debugged(OpName)),nl,
   findall(TP, (checked_property(P),
               translate_bexpression_with_limit(P,TrP),
               ajoin([TrP,' &\n'],TP)),
               Res1),
   ((projection_onto_component(_Nr))
     -> Res1P = ['(only a sub-component was analysed)\n'|Res1] ; Res1P=Res1),
   (Res1=[] -> R1=Res1P ;
     R1=['\n ********\n\nSATISFIABLE CONSTRAINTS and PROPERTIES:\n'|Res1P]
   ),
   findall(TP2,( ((failed_property(P,_),RES='false') ; (time_out_property(P,_),RES='time_out')),
               translate_status(P,RES,TP2)),
               Res2),
   (Res2=[] -> R2=['\nNo unsatisfiable PROPERTY or CONSTRAINT'], Satisfiable=true
           ;   R2=['\n ********\n\nFirst unsatisfiable PROPERTY or CONSTRAINT:\n'|Res2],
               Satisfiable=false
   ),
   findall(TP, (last_checked_solution(ConstantsState,_), % TO DO: Project on remaining variables !?
               translate_equivalence(b(identifier('SOLUTION'),boolean,[]),ConstantsState,TP)),
               Res12),
   append(R2,['\n\n ********\n\nSolution for satisfiable part:\n'|Res12],R2Res12),
   append(R1,R2Res12,Res),
   findall(SetDescr,(b_fd_type(Set,LowBnd,UpBnd), Card is 1+UpBnd-LowBnd,
             create_texpr(identifier(Set),set(global(Set)),[],SetId),
             create_texpr(card(SetId),integer,[],CardExpr),
             create_texpr(value(int(Card)),integer,[],ValExpr),
             create_texpr(equal(CardExpr,ValExpr),prop,[],EqualExpr),
             translate_bexpression(EqualExpr,SetDescr1),
             (b_global_sets:unfixed_deferred_set(Set) -> Xtra = ' (assumed for deferred set)' ; Xtra=''),
             ajoin([SetDescr1,Xtra,'\n'],SetDescr)), Sets),

   append(['\nCardinalities of SETS:\n'|Sets],Res,SRes),
   preferences:get_preference(minint,MININT),
   preferences:get_preference(maxint,MAXINT),
   clear_error_context,
   ajoin(['MININT .. MAXINT = ',MININT,'..',MAXINT],MINMAX),
   RealRes = ['SETTINGS:\n',MINMAX|SRes].
tcltk_debug_properties_or_op([],_,list(['Machine has no PROPERTIES']),true).

failed_or_time_out_property(P) :- failed_property(P,_) ; time_out_property(P,_).

:- use_module(library(codesio),[write_to_codes/2]).
:- use_module(bsyntaxtree,[find_typed_identifier_uses/2]).
:- use_module(tools,[ajoin_with_sep/3, ajoin/2]).
% use_module(probsrc(predicate_debugger)),predicate_debugger:print_unsat_core_with_types
tcltk_get_unsat_core_with_types(ResultCodes) :-
   get_unsat_core(AllConj),
   translate_predicate_into_machine(AllConj,ResultAtom),
   write_to_codes(ResultAtom,ResultCodes).

get_unsat_core(AllConj) :- (checked_property(_) ; failed_or_time_out_property(_)), !,
   findall(P,checked_property(P),Ps),
   findall(UnsatP,failed_or_time_out_property(UnsatP),UPs),
   append(Ps,UPs,All),
   conjunct_predicates(All,AllConj).
get_unsat_core(Core) :- print('No predicate debugger analysis run'),nl,
   print('Trying to extract from b_interpreter'),nl,
   b_interpreter:get_unsat_component_predicate(Core,_Vars),!.
get_unsat_core(b(truth,pred,[])).


translate_equivalence(A,B,R) :-
   translate_bexpression(A,TA),
   translate_bstate(B,TB),
   ajoin([TA,'\n   == ',TB,'\n'],R).
:- use_module(translate,[translate_span/2]).
translate_status(A,Status,R) :-
   translate_bexpression(A,TA),
   (translate_span(A,ASpan)
    ->  ajoin([TA,'\n   == ',Status,'   (line:col) =',ASpan],R)
     ;  ajoin([TA,'\n   == ',Status],R)).



:- use_module(bmachine).
b_debug_properties :- set_error_context(b_debug_properties),
   reset_debug_properties,
   %%b_extract_constants_and_properties(Constants,CstTypes,Properties),
   %b_machine_has_constants_or_properties,
   b_get_machine_constants(TypedConstants),
   b_get_properties_from_machine(Properties),!,
   adapt_properties(Properties,TypedConstants,AProperties,ATypedConstants,AllowTimeout),
   nr_of_conjuncts(AProperties,NrConj),
   (b_debug_properties2(AProperties,ATypedConstants,[],NrConj,AllowTimeout) -> true ; true).

b_debug_goal :-
    bmachine:b_get_machine_goal(Goal),
    b_debug_predicate(Goal).
b_debug_predicate(FullPred) :- set_error_context(b_debug_predicate),
   reset_debug_properties,
   get_parameters(FullPred,Params,Pred),
   get_current_state('@PRED',CurID,CurBState),
   set_error_context(b_debug_predicate(CurID)),
   nr_of_conjuncts(Pred,NrConj),
   (b_debug_properties2(Pred,Params,CurBState,NrConj,allow_timeout) -> true ; true).

get_parameters(b(exists(Parameters,Pred),pred,_),Parameters,Pred) :- !.
get_parameters(Pred,[],Pred).

:- use_module(library(lists)).
adapt_properties(Properties,TypedConstants,Pred,CTypedConstants,do_not_allow_timeout) :-
    %(b_interpreter:nr_of_components(NrC), NrC>1),
    preferences:get_preference(partition_predicates,true),
    b_interpreter:unsat_properties_component(CompNr),!, % means there was no timeout
    assert(projection_onto_component(CompNr)),
    print(debugging_unsat_or_skipped_component(CompNr)),nl,
    bsyntaxtree:predicate_components(Properties,Components),
    nth1(CompNr,Components,component(Pred,Vars)),
    print(component_vars(Vars)),nl,
    project_texpr_identifiers(TypedConstants,Vars,CTypedConstants).
adapt_properties(Properties,TypedConstants,Properties,TypedConstants,allow_timeout). % TO DO determine whether timeout occured


% only keep texpr_identifiers whose id is in the provided list
project_texpr_identifiers([],_,[]).
project_texpr_identifiers([TID|T],Vars,Res) :-
   (get_texpr_id(TID,ID),member(ID,Vars)
    -> Res = [TID|R2] ; Res=R2),
   project_texpr_identifiers(T,Vars,R2).

:- use_module(specfile,[state_corresponds_to_initialised_b_machine/2, state_corresponds_to_set_up_constants/2]).
/* b_interpreter:tcltk_debug_properties_or_op(prog1,R)*/
b_debug_operation(OpName) :-
   reset_debug_properties,
   get_current_state(OpName,CurID,CurBState),
   set_error_context(operation(OpName,CurID)),
   (get_operation_enabling_condition(OpName,Parameters,EnablingCondition,BSVars,_,true,false) -> true
     ; add_error(b_debug_operation,'Operation does not exist or cannot determine enabling condition: ',OpName),fail),!,
   print('Enabling: '), print_bexpr(EnablingCondition),nl,
   flush_output(user),
   nr_of_conjuncts(EnablingCondition,NrConj),
   ~~mnf(insert_before_substitution_variables(BSVars,[],CurBState,CurBState,NewCurBState)),
   extract_exists_parameters(EnablingCondition,Parameters,NewCond,NewParas),
   (b_debug_properties2(NewCond,NewParas,NewCurBState,NrConj,allow_timeout) -> true ; true).

:- use_module(state_space,[current_expression/2]).
get_current_state(OpName,CurID,CurBState) :-
   (current_expression(CurID,CurState) -> true
      ; add_error(predicate_debugger,'Could not get current state',OpName),fail),
    (state_corresponds_to_initialised_b_machine(CurState,CurBState) -> true
     ; (OpName=='$initialise_machine',
        state_corresponds_to_set_up_constants(CurState,CurBState)) -> true
     ; add_error(b_debug_operation,'Machine not yet initialised! ',CurState),fail).


 % Try to lift out existentially quantified variables to the outer level; to make the debugging more fine-grained
extract_exists_parameters(Pred,Parameters, NewPred, NewParas) :-
    get_texpr_expr(Pred,exists(Vars,Body)),
    disjoint_list_union(Parameters,Vars,Paras1), !,  % also checks if no clash exists; otherwise we have to keep exists
    print(augmented_paras(Paras1)),nl,
    extract_exists_parameters(Body,Paras1,NewPred,NewParas).
extract_exists_parameters(Pred,Parameters, NewPred, NewParas) :-
    get_texpr_expr(Pred,conjunct(LHS,RHS)),!,
    extract_exists_parameters(LHS,Parameters, NewLHS, IntParas),
    extract_exists_parameters(RHS,IntParas,NewRHS,NewParas),
    conjunct_predicates([NewLHS,NewRHS],NewPred).
extract_exists_parameters(Pred,Par,Pred,Par).


:- dynamic obtain_action_enabling_conditions/0, simplify_enabling_condition/1.
obtain_action_enabling_conditions. simplify_enabling_condition(true).
set_obtain_action_enabling_conditions(true) :- !,
 (obtain_action_enabling_conditions -> true ; assert(obtain_action_enabling_conditions)).
set_obtain_action_enabling_conditions(_) :- retractall(obtain_action_enabling_conditions).
set_simplify_enabling_condition(X) :- retractall(simplify_enabling_condition(_)),
  assert(simplify_enabling_condition(X)).

:- use_module(b_interpreter_components,[construct_optimized_exists/3,construct_optimized_exists/4]).
:- use_module(specfile,[animation_minor_mode/1]).
:- use_module(external_functions,[external_subst_enabling_condition/3]).

get_unsimplified_operation_guard(OpName,Guard) :-
    get_unsimplified_operation_enabling_condition(OpName,Parameters,EnablingCondition,_BecomesSuchVars,_Precise),
    construct_optimized_exists(Parameters,EnablingCondition,Guard).

get_unsimplified_operation_enabling_condition(OpName,Parameters,EnablingCondition,BecomesSuchVars,Precise) :-
    Simplify=false,
    (animation_minor_mode(eventb) -> GetFromAction=false
     ; GetFromAction=true),
    get_operation_enabling_condition(OpName,Parameters,EnablingCondition,BecomesSuchVars,Precise,GetFromAction,Simplify).

get_simplified_operation_enabling_condition(OpName,Parameters,EnablingCondition,BecomesSuchVars,Precise) :-
    Simplify=true,
    (animation_minor_mode(eventb) -> GetFromAction=false
     ; GetFromAction=true),
    get_operation_enabling_condition(OpName,Parameters,EnablingCondition,BecomesSuchVars,Precise,GetFromAction,Simplify).

get_operation_enabling_condition(OpName,Parameters,EnablingCondition,BecomesSuchVars,IsPrecise,GetAct,Simplify) :-
   set_obtain_action_enabling_conditions(GetAct), % specify whether we want also conditions stemming from actions, such as :() or ::
   set_simplify_enabling_condition(Simplify), % specify whether we want to simplify the EnablingCondition if possible
   %%nonvar(OpName),  %% I am not sure why this test is here ??
   %% print(getting_enabling(OpName,GetAct)),nl,%%
   b_get_machine_operation_for_animation(OpName,_Results,Parameters,TBody,_OType,true),
   get_operation_enabling_condition1(TBody,EnablingCondition,BecomesSuchVars,IsPreciseV),
   (var(IsPreciseV) -> IsPrecise = precise ; IsPrecise = imprecise).

get_operation_enabling_condition1(TBody,EnablingCondition,BecomesSuchVars,IsPrecise) :-
   get_texpr_expr(TBody,Body),
   (get_operation_enabling_condition2(Body,EnablingCondition,BecomesSuchVars,IsPrecise) -> true
     ; add_internal_error('Getting enabling condition failed: ',Body),
       %trace,get_operation_enabling_condition2(Body,EnablingCondition,BecomesSuchVars),
       create_texpr(truth,pred,[],EnablingCondition),
       BecomesSuchVars=[]).

 /* TO DO: recursively expand the code below for more complicated SELECT,ANY,... */
get_operation_enabling_condition2(precondition(PreCond,TBody),Res,BecomesSuchVars,IsPrecise) :- !, % PRE
     get_operation_enabling_condition1(TBody,RC,BecomesSuchVars,IsPrecise),
     ~~mnf(conjunct_predicates([PreCond,RC],Res)).
get_operation_enabling_condition2(assertion(_PreCond,TBody),Res,BecomesSuchVars,IsPrecise) :- !, % ASSERT
     get_operation_enabling_condition1(TBody,Res,BecomesSuchVars,IsPrecise).
get_operation_enabling_condition2(block(TBody),Res,BecomesSuchVars,IsPrecise) :- !, % BEGIN  ... END
     get_operation_enabling_condition1(TBody,Res,BecomesSuchVars,IsPrecise).
get_operation_enabling_condition2(choice(ChoiceList),Res,BecomesSuchVars,IsPrecise) :- !, % CHOICE ... OR ...
     get_disjunction_of_enabling_conditions(ChoiceList,Res,BecomesSuchVars,IsPrecise).
get_operation_enabling_condition2(var(_Parameters,TBody),Res,BecomesSuchVars,IsPrecise) :- !, % VAR
     % should the Parameters be added somewhere ? In principle the enabling condition should always be true, as we have a low-level construct
     get_operation_enabling_condition1(TBody,Res,BecomesSuchVars,IsPrecise).
get_operation_enabling_condition2(sequence([TBody1|_]),Res,BecomesSuchVars,imprecise) :- !, % ;
     (debug:debug_mode(on) -> print(ignoring_potential_enabling_condition_in_tail_of_sequence),nl ; true),
     % TO DO: compute before-after predicate of TBody1 ? before_after_predicate_list_conjunct_with_equalities ?
     % we ignore rest of sequence; cannot influence enabling condition; probably this should always return truth
     get_operation_enabling_condition1(TBody1,Res,BecomesSuchVars,_IsPrecise).
get_operation_enabling_condition2(lazy_let_subst(ID,ExprOrPred,TBody),Res,BecomesSuchVars,IsPrecise) :- !,
     Res = b(lazy_let_pred(ID,ExprOrPred,BodyRes),pred,[]),
     get_operation_enabling_condition1(TBody,BodyRes,BecomesSuchVars,IsPrecise).
get_operation_enabling_condition2(let(Parameters,PreCond,TBody),Res,BecomesSuchVars,IsPrecise) :- !, % LET
     % for the moment: treat like ANY, but we could avoid introducing existential quantifier ?!
     get_operation_enabling_condition2(any(Parameters,PreCond,TBody),Res,BecomesSuchVars,IsPrecise).
get_operation_enabling_condition2(any(Parameters,PreCond,TBody),Res,BecomesSuchVars,IsPrecise) :- !, % ANY
     get_operation_enabling_condition1(TBody,RC,BecomesSuchVars,IsPrecise),
     ~~mnf(conjunct_predicates([PreCond,RC],Res1)),
     simplify_enabling_condition(Simplify),
     construct_optimized_exists(Parameters,Res1,Res,Simplify).
%get_operation_enabling_condition2(select([b(select_when(PreCond, TBody),_,_)]),Res,BecomesSuchVars) :- !,
%     get_texpr_expr(TBody,Body),
%     ~~mnf(get_operation_enabling_condition2(Body,RC,BecomesSuchVars)),
%     ~~mnf(conjunct_predicates([PreCond,RC],Res)).
%get_operation_enabling_condition2(select([b(select_when(PreCond, TBody),_,_)],b(skip,_,_)),Res,BecomesSuchVars) :- !,
%     get_texpr_expr(TBody,Body),
%     ~~mnf(get_operation_enabling_condition2(Body,RC,BecomesSuchVars)),
%     ~~mnf(conjunct_predicates([PreCond,RC],Res)).
get_operation_enabling_condition2(select_when(PreCond,TBody),Res,BecomesSuchVars,IsPrecise) :- !,
     get_operation_enabling_condition1(TBody,RC,BecomesSuchVars,IsPrecise),
     ~~mnf(conjunct_predicates([PreCond,RC],Res)).
get_operation_enabling_condition2(select(ListOfWhens),Res,BecomesSuchVars,IsPrecise) :- !,
     get_disjunction_of_enabling_conditions(ListOfWhens,Res,BecomesSuchVars,IsPrecise).
get_operation_enabling_condition2(select(ListOfWhens,Else),Res,BecomesSuchVars,IsPrecise) :- !,
     get_texpr_exprs(ListOfWhens,ListOfSelectWhens),
     maplist(get_operation_enabling_condition3(IsPrecise),ListOfSelectWhens,Res1,BecomesSuchVars1),
     get_operation_enabling_condition1(Else,ResElse,ElseBecomesSuchVars,IsPrecise),
     ~~mnf(disjunct_predicates([ResElse|Res1],Res)),
     append([ElseBecomesSuchVars|BecomesSuchVars1],BecomesSuchVars).
get_operation_enabling_condition2(parallel([TH|T]),Res,BecomesSuchVars,IsPrecise) :- !,
     get_operation_enabling_condition1(TH,E1,BecomesSuchVarsH,IsPrecise),
     (T=[] -> Res=E1, BecomesSuchVarsH=BecomesSuchVars
       ; (~~mnf(get_operation_enabling_condition2(parallel(T),E2,BecomesSuchVarsT,IsPrecise)),
          append(BecomesSuchVarsH,BecomesSuchVarsT,BecomesSuchVars),
          conjunct_predicates([E1,E2],Res))).
get_operation_enabling_condition2(becomes_element_of(_LHS,RHS),Res,[],_Precise) :-
     obtain_action_enabling_conditions,
     get_texpr_expr(RHS,RHSExpr),get_texpr_type(RHS,Type),!,
     ((simplify_enabling_condition(true),definitely_not_empty(RHSExpr,Type))
       -> %print(def_not_empty(RHSExpr)),nl,
          create_texpr(truth,pred,[],Res)
       ;  create_texpr(empty_set,Type,[],EmptySet),
          safe_create_texpr(not_equal(RHS,EmptySet),pred,Res2),
          %print(not_equal_empty(RHSExpr,Type)),nl,
          ~~mnf(clean_up(Res2,[],Res))
     ).
get_operation_enabling_condition2(becomes_such(Vars,Condition),Res,Vars,_Precise) :-  %  Vars : (Condition)
     % example x: (x$0 >= 1 & x=x$0+1)
     obtain_action_enabling_conditions,!,
     simplify_enabling_condition(Simplify),
     construct_optimized_exists(Vars,Condition,Res1,Simplify),
     %translate:print_bexpr(Res1),nl,
     % e.g, for example above we have #x.(x$0 >= 1 & x=x$0+1) -> x$0 >= 1
     % now rename $0 variables to act on current state to obtain the condition:
     findall(rename(BeforeId,Id),
             (member(b(identifier(Id),_,Infos),Vars),
              member(before_substitution(_,BeforeId),Infos)),
              RenameList),
     rename_bt(Res1,RenameList,Res). % for example above: Res is x >= 1
get_operation_enabling_condition2(rlevent(_Name,_Section,_Status,_Params,Guard,_Theorems,Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents),Res,BecomesSuchVars,IsPrecise) :- !,
   %print(actions(Actions)),nl,
   % TO DO: have a look at get_full_eventb_guard. Do we want to recurse through the abstractions ?
   (obtain_action_enabling_conditions
    ->  get_operation_enabling_for_event_b_actions(Actions,Guard,Res,BecomesSuchVars,IsPrecise)
    ;  Res=Guard,BecomesSuchVars=[]).
get_operation_enabling_condition2(assign(_LHS,_RHS),Truth,[],_Precise) :- !,
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(assign_single_id(_IDE,_RHS),Truth,[],_Precise) :- !,
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(case(A,_,_),Truth,[],imprecise) :- !, % CASE is now translated to LET + IF-THEN-ELSE in b_ast_cleanup
     print('Not computing enabling for CASE: '), translate:print_bexpr(A),nl,
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(while(_,_,_,_),Truth,[],imprecise) :- !,
     debug:debug_println(9,enabling_for_while_assumed_true),
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(if(IfList),Res,BecomesSuchVars,IsPrecise) :- !,
     maplist5(get_if_condition(IsPrecise),IfList,Tests,Conds,BecomesSuchVars1),
     append(BecomesSuchVars1,BecomesSuchVars),
     ((member(X,Conds), \+ is_truth(X))
       -> disjoin_ifs(Tests,Conds,[],L),
          conjunct_predicates(L,Res)
       ; create_texpr(truth,pred,[],Res)  % all branches have no enabling condition
      ).
get_operation_enabling_condition2(skip,Truth,[],_Prcise) :- !,
     create_texpr(truth,pred,[],Truth).
get_operation_enabling_condition2(external_subst_call(Pred,Args),Res,[],_Precise) :-
     external_subst_enabling_condition(Pred,Args,Cond),!, Res=Cond.
get_operation_enabling_condition2(operation_call(Operation,_OpCallResults,OpCallParas),Res,BecomeSuchVars,IsPrecise) :-
    def_get_texpr_id(Operation,op(OperationName)),
    b_get_machine_operation_for_animation(OperationName,_OpResults,OpParameters,Body,_OType,_TopLevel),
    bsyntaxtree:replace_ids_by_exprs(Body,OpParameters,OpCallParas,Body2),
    !,
    %print(get_enabling_for_opcall),nl,translate:print_subst(Body2),nl,
    get_operation_enabling_condition1(Body2,Res,BecomeSuchVars,IsPrecise).
get_operation_enabling_condition2(X,Truth,[],imprecise) :- %print(cannot_obtain(X)),nl,
     (obtain_action_enabling_conditions
      -> functor(X,F,A),debug_println(9,cannot_obtain_enabling(F/A,X)) ; true),
     create_texpr(truth,pred,[],Truth).


get_if_condition(IsPrecise,b(if_elsif(Test,TBody),_,_),Test,Condition,BecomesSuchVars) :-
     get_operation_enabling_condition1(TBody,Condition,BecomesSuchVars,IsPrecise).

:- use_module(bsyntaxtree, [create_implication/3]).
disjoin_ifs([],[],_,[]).
disjoin_ifs([Test|TT],[EnableCond|TC],NegSoFar,[Res1|TR]) :-
   append(NegSoFar,[Test],L),
   conjunct_predicates(L,BranchTest),
   create_implication(BranchTest,EnableCond,Res1),  %  ELSEIF Test THEN Body ...  ---> NegSoFar & Test => EnableCond
   create_negation(Test,NTest),
   append(NegSoFar,[NTest],NegSoFar1), % add negation of test as additional test for rest
   disjoin_ifs(TT,TC,NegSoFar1,TR).


% for a list of substitutions: get enabling conditions and disjoin them
get_disjunction_of_enabling_conditions(ListOfWhens,Res,BecomesSuchVars,IsPrecise) :-
     get_texpr_exprs(ListOfWhens,ListOfSelectWhens),
     maplist(get_operation_enabling_condition3(IsPrecise),ListOfSelectWhens,Res1,BecomesSuchVars1),
     ~~mnf(disjunct_predicates(Res1,Res)),
     append(BecomesSuchVars1,BecomesSuchVars).
get_operation_enabling_condition3(Precise,Exp,Res,BV) :- get_operation_enabling_condition2(Exp,Res,BV,Precise).

definitely_not_empty(bool_set,_).
definitely_not_empty(integer_set(_),_).
definitely_not_empty(identifier(X),set(global(X))).
definitely_not_empty(string_set,_).



% get operation enabling condition in context of an event-b action list:
get_operation_enabling_for_event_b_actions([],Res,Res,[],_Precise).
get_operation_enabling_for_event_b_actions([H|T],InRes,OutRes,BecomesSuchVars,IsPrecise) :-
     get_operation_enabling_condition1(H,ResH,HBecomesSuchVars,IsPrecise),
     ~~mnf(conjunct_predicates([InRes,ResH],IntRes)),
     get_operation_enabling_for_event_b_actions(T,IntRes,OutRes,TBecomesSuchVars,IsPrecise),
     append(HBecomesSuchVars,TBecomesSuchVars,BecomesSuchVars).

% ---------------
:- use_module(bmachine,[b_top_level_operation/1]).
% LTSMIN style guards: a guard that does not depend on parameters
get_operation_propositional_guards(OpName,Guards,RestBody) :-
   b_top_level_operation(OpName),
   b_get_machine_operation_for_animation(OpName,TResults,TParameters,TBody,_OType,true), % requires bmachine to be precompiled
   get_operation_propositional_guards(OpName,TResults,TParameters,TBody,Guards,RestBody).

:- use_module(translate,[print_bexpr/1]).
% the following can be called directly; does not require bmachine to be pre-compiled
get_operation_propositional_guards(OpName,TResults,TParameters,TBody,Guards,RestBody) :-
   get_texpr_ids(TParameters,Ids1),
   get_texpr_ids(TResults,Ids2),
   append(Ids1,Ids2,Ids), sort(Ids,Parameters),
   get_operation_guards_aux(TBody,Parameters,top,Guards,RestBody),
   (debug:debug_mode(off) -> true
     ; format('OPERATION Guard Splitting ~w (~w)~n',[OpName,Parameters]),
       print('LTSMin Guards: '), maplist(print_bexpr,Guards),nl,
       print('LTSMin Body: '), translate:print_subst(RestBody),nl,nl
    ).

:- use_module(bsyntaxtree, [conjunction_to_list/2,find_identifier_uses/3]).


get_operation_guards_aux(Subst,Parameters,Top,Guards,OpBody) :-
   get_guards(Subst,Top,TIds,Guard,InnerBody,Infos),
   get_texpr_ids(TIds,Ids), sort(Ids,NewIds),
   ord_union(Parameters,NewIds,Parameters2),
   !,
   get_parameter_independent_guards(Guard,Parameters2,Indep,Dep),
   (Dep = []
    -> OpBody = InnerOpBody  % no need to keep b
    ; conjunct_predicates(Dep,DepCond),
      construct_select(TIds,DepCond,InnerOpBody,Infos,OpBody)  % we always produce a SELECT; even if we had a PRE as it will no longer be innermost ! we assume treat_outermost_pre_as_select is set to true for PRE (checked below)
    ),
   get_operation_guards_aux(InnerBody,inner,Parameters2,InnerGuards,InnerOpBody),
   append(Indep,InnerGuards,Guards).

get_operation_guards_aux(b(rlevent(Name,Sect,Status,Params,EvGuard,Theorems,Act,VWit,PWit,Unmod,AbsEvts),subst,Info),
                     Parameters,_,Guards,OpBody) :- !,
   get_parameter_independent_guards(EvGuard,Parameters,InDepGuards,Dep),
   (get_variant_pred(Status,VariantPred)
     -> %print(op_variant(_Name)),nl,translate:print_bexpr(VariantPred),nl,
        % we virtually include the Variant expression in the read info; ensure that LTS Min knows that this will be read by the Event-B interpreter
        % TO DO: check if we need to add the witnesses as well !
        Guards = [VariantPred|InDepGuards]
     ; Guards = InDepGuards),
   conjunct_predicates(Dep,DepG),
   OpBody = b(rlevent(Name,Sect,Status,Params,DepG,Theorems,Act,VWit,PWit,Unmod,AbsEvts),subst,Info).
get_operation_guards_aux(TB,_,_,[],TB).


:- use_module(bsyntaxtree, [safe_create_texpr/3]).
% we integrate the Variant check into the guard to ensure the correct read matrix is produced
% currently the Variant is checked upon event entry and for convergent events upon exit for decrease
% in principle this should be more of an assertion_predicate or assertion_expression
get_variant_pred(b(Status,status,_),Res) :- get_variant_pred_aux(Status,Res).
get_variant_pred_aux(anticipated(Variant),Res) :- NATURAL = b(integer_set('NATURAL'),set(integer),[]),
   safe_create_texpr(member(Variant,NATURAL),pred,Res).
get_variant_pred_aux(convergent(Variant),Res) :- NATURAL1 = b(integer_set('NATURAL1'),set(integer),[]),
   safe_create_texpr(member(Variant,NATURAL1),pred,Res).


get_guards(b(precondition(Guard,TBody),subst,Info), top, [],Guard, TBody, Info) :-
   preferences:get_preference(treat_outermost_pre_as_select,true).
get_guards(b(select([b(select_when(Guard, TBody),subst,_Info1)]),subst,Info2), _, [],Guard, TBody, Info2).
% TO DO: for ANY try and extract propositional parts, e.g., for ANY pp WHERE pp:1..xx & z=1/yy THEN …
get_guards(b(any(TIds,Guard,TBody),subst,Info),_, TIds,Guard,TBody,Info).


construct_select([],Guard,TBody, Infos, Res) :- !,
    Res = b(select([b(select_when(Guard, TBody),subst,Infos)]),subst,Infos).
construct_select(TIds,Guard,TBody,Infos,  b(any(TIds,Guard,TBody),subst,Infos)).


get_parameter_independent_guards(Guard,Parameters,Indep,Dep) :-
   conjunction_to_list(Guard,Gs),
   l_get_parameter_independent_guards(Gs,Parameters,at_front,Indep,Dep).

:- use_module(library(ordsets)).
:- use_module(bsyntaxtree,[always_well_defined/1]).
l_get_parameter_independent_guards([],_,_,[],[]).
l_get_parameter_independent_guards([G|Gs],Parameters,AtFront,Indep,Dep) :-
   find_identifier_uses(G,[],Ids),
   ((ord_disjoint(Ids,Parameters),
     (AtFront=at_front -> true ; always_well_defined(G)))
    -> Indep=[G|I1], Dep=D1, AtFront1=AtFront
     ; Indep=I1, Dep=[G|D1],
       AtFront1=not_at_front % we have skipped one guard; the next guard is not guaranteed to be at the front (relevant for well-definedness)
   ),
   l_get_parameter_independent_guards(Gs,Parameters,AtFront1,I1,D1).

% -----------------------------------------

:- use_module(library(timeout)).


nr_of_conjuncts(Pred,Nr) :-
  is_a_conjunct(Pred,LHS,RHS),!,
  nr_of_conjuncts(LHS,NrL),
  nr_of_conjuncts(RHS,NrR),
  Nr is NrL+NrR.
nr_of_conjuncts(_,1).

:- volatile last_debug_info/2.
:- dynamic last_debug_info/2.
b_debug_properties2(Pred,Constants,GlobalState,TotalNr,AllowTimeout) :-
  create_texpr(truth,pred,[b_debug_properties],Truth),
  retractall(last_debug_info(_,_)), assert(last_debug_info(Constants,GlobalState)),
  (b_debug_find_first_unsat_property(Pred,Truth,Constants,GlobalState,TotalNr,AllowTimeout)
  -> print('ALL PROPERTIES SATISFIABLE'),nl
  ;  true % get_preference(debug_try_minimise,false) -> true ;  b_find_unsat_core(Constants,GlobalState)
  ).

b_find_unsat_core :-     /* assumes that b_debug_properties was run before */
  % b_machine_has_constants_or_properties,
 %  b_get_machine_constants(TypedConstants),
  % b_get_properties_from_machine(Properties),!,
  % adapt_properties(Properties,TypedConstants,_AProperties,ATypedConstants),
   %b_find_unsat_core(ATypedConstants,[]).
   last_debug_info(Constants,GlobalState), b_find_unsat_core(Constants,GlobalState).
b_find_unsat_core(Constants,GlobalState) :-
     print('TRYING TO MINIMISE UNSATISFIABLE PROPERTIES'),nl,
     (last_checked_solution(_,Time)
      -> print(' Time for last solution in ms: '), print(Time), nl ; true),
     findall(P,checked_property(P),SatProps),
     (failed_property(FailProp,USTime) -> true
      ; time_out_property(FailProp,USTime) -> true
      ; add_error_fail(b_find_unsat_core,'No failed or timed-out property:',SatProps)),
     print(' Time for unsatisfiable conjuncts in ms: '), print(USTime), nl,
     append(SatProps,[FailProp],Props),
     length(SatProps,SL),
     b_debug_try_remove_props(SatProps,Props,[],Constants,GlobalState,1,SL).
     % TO DO: partition into components; try removing compnents without FailProp

b_debug_try_remove_props([],_,_,_,_,_,_) :- !.
b_debug_try_remove_props([],_,_,TypedConstants,GlobalState,Count,_) :- !, % for debug
  format('~nFINISHED (~w); CHECKING CORE~n',[Count]),
  get_unsat_core(Core),
  nl,translate:print_bexpr(Core),nl,
  (timeout_fast_check_property(unsat,Core,TypedConstants,_,GlobalState,TimeOutResC) -> true
   ; TimeOutResC = 'UNSAT'),
  format('~nCORE RESULT: ~w~n',[TimeOutResC]).
b_debug_try_remove_props([H|TSat],[H|TUnsat],Required,TypedConstants,GlobalState,Count,Total) :-
  append(Required,TSat,ConjunctsWithoutH),
  conjunct_predicates(ConjunctsWithoutH,SatPred),
  translate_bexpression(H,PT),
  format('~nCHECKING REMOVAL(~w/~w) OF: ~w~n',[Count,Total,PT]),
  (timeout_fast_check_property(sat,SatPred,TypedConstants,Sol,GlobalState,TimeOutResSat)
   -> (\+ is_time_out_result(TimeOutResSat)
       -> append(Required,TUnsat,ConjunctsWithoutH2),
          conjunct_predicates(ConjunctsWithoutH2,UnsatPred),
          ((timeout_fast_check_property(unsat,UnsatPred,TypedConstants,_,GlobalState,TimeOutRes),
            print(res(TimeOutRes)),nl,
            (is_time_out_result(TimeOutRes) ->  \+ time_out_property(_,_) ; true)
           )
               -> print('Required for inconsistency : '), print(TimeOutRes),nl,
                   append(Required,[H],NewRequired)
               ;  /* H is not required for inconsistency/timeout */
                   print('NOT required for INCONSISTENCY !'),nl,
                   % print_bexpr(UnsatPred),nl,
                   Required=NewRequired,
                   % TO DO: update found solution
                   (retract(last_checked_solution(_,LastTime)) -> true
                     ; print('*** Could not remove last_checked_solution'),nl,
                       preferences:get_computed_preference(debug_time_out,LastTime)
                   ),
                   assert(last_checked_solution(Sol,LastTime)),
                   (retract(checked_property(H)) -> true
                    ; print('*** Could not remove checked_property: '), print(H),nl
                   )
          )
      ; print('Required for efficiency'),nl,
        append(Required,[H],NewRequired)
      )
  ;  nl,print('*** INTERNAL ERROR: FAILED'),nl,nl,
     append(Required,[H],NewRequired)
  ),
  C1 is Count+1,
  b_debug_try_remove_props(TSat,TUnsat,NewRequired,TypedConstants,GlobalState,C1,Total).

b_debug_find_first_unsat_property(Pred,Rest,Constants,GlobalState,TotalNr,AllowTimeout) :-
  is_a_conjunct(Pred,LHS,RHS),!,
  (b_debug_find_first_unsat_property(LHS,Rest,Constants,GlobalState,TotalNr,AllowTimeout)
   -> conjunct_predicates([Rest,LHS],Rest2),
      b_debug_find_first_unsat_property(RHS,Rest2,Constants,GlobalState,TotalNr,AllowTimeout)
   ; fail
  ).
b_debug_find_first_unsat_property(Prop,Rest,TypedConstants,GlobalState,TotalNr,AllowTimeout) :-
  nl,nl,flush_output(user),
  conjunct_predicates([Rest,Prop],Pred),
  (retract(cur_conjunct_nr(CurNr)) -> true ; CurNr=1), CurNrP1 is CurNr+1,
  assert(cur_conjunct_nr(CurNrP1)),
  translate_bexpression(Prop,PT),
  format('~nCHECKING (~w/~w) OF: ~w~n',[CurNr,TotalNr,PT]),
    statistics(runtime,[Start,_]),
  (timeout_check_property(Pred,TypedConstants,ConstantsState,GlobalState,TimeOutRes)
   -> statistics(runtime,[Stop,_]), Time is Stop - Start,
      print_time(TimeOutRes,Time),
      (((\+ last_checked_solution(_,_)  % we haven't found a solution yet
          ; AllowTimeout=allow_timeout),
        is_time_out_result(TimeOutRes))
         -> assert(time_out_property(Prop,Time))
         ;
            assert_checked_property(Prop,ConstantsState,Time)
      )
   ;  statistics(runtime,[Stop,_]), Time is Stop - Start,
      assert(failed_property(Prop,Time)),
      print('No Solution Found !! ('),print(Time), print(' ms)'),nl, flush_output(user),
      %print_bexpr(Pred),nl,
      fail
  ).

print_time(TimeOutRes,Time) :- is_time_out_result(TimeOutRes),!,
   (TimeOutRes=virtual_time_out(_) -> print('VIRTUAL ') ; true),
    print('TIMEOUT !! ('),print(Time), print(' ms)'),nl, flush_output(user).
print_time(_TimeOutRes,Time) :-
    print('OK ('), print(Time), print(' ms)'), flush_output(user).


last_unsat_time(Time) :- time_out_property(_,Time) -> true ; failed_property(_,Time).

set_fast_unsat_core(T) :- retractall(fast_unsat_core),
  (T==fast -> assert(fast_unsat_core) ; true).
:- dynamic fast_unsat_core/0.
fast_unsat_core.

additional_time(301). % additional time given to account for variations in solving
%additional_time(1301).

get_timeout(sat,DebugTimeOut) :-
   last_checked_solution(_,LTime),
   !,
   additional_time(AT),
   DebugTimeOut is LTime+AT.
get_timeout(unsat,DebugTimeOut) :-
   last_unsat_time(USTime),
   last_checked_solution(_,LSTime),
    (fast_unsat_core -> TT is min(USTime,LSTime) ; TT is max(USTime,LSTime)),
    !,
   additional_time(AT),
    DebugTimeOut is TT+AT.
get_timeout(S,DebugTimeOut) :- print('*** Could not get TimeOut : '), print(S),nl,
   preferences:get_preference(time_out,CurTO), DebugTimeOut is CurTO * 1.


timeout_fast_check_property(ExpectSat,Pred,TypedConstants,ConstantsState,GlobalState,TimeOutRes) :-
   get_timeout(ExpectSat,DebugTimeOut),
   % print(time_out_value(DebugTimeOut)),nl,
   % print_bexpr(Pred),nl,
  statistics(runtime,[Start,_]),
  time_out_with_enum_warning_one_solution(b_check_property_satisfiable(Pred,TypedConstants,ConstantsState,GlobalState),
            DebugTimeOut,TimeOutRes),
  statistics(runtime,[Stop,_]), Time is Stop - Start, print(time(Time)),nl,
  (is_time_out_result(TimeOutRes) -> print(time_out_occurred(Time,DebugTimeOut)),nl,
    (debug:debug_mode(on) -> print_bexpr(Pred),nl ; true)
   ; true).

timeout_check_property(Pred,TypedConstants,ConstantsState,GlobalState,TimeOutRes) :-
  preferences:get_computed_preference(debug_time_out,DebugTimeOut),
  time_out_with_enum_warning_one_solution(b_check_property_satisfiable(Pred,TypedConstants,ConstantsState,GlobalState),
            DebugTimeOut,TimeOutRes),
    format('~nCHECKING RESULT for: ~w~n',[TimeOutRes]).

assert_checked_property(P,ConstantsState,Time) :-
   assert(checked_property(P)),
   (var(ConstantsState) -> print('NO SOLUTION'),nl
    ; retractall(last_checked_solution(_,_)),
      assert(last_checked_solution(ConstantsState,Time))).

:- use_module(kernel_waitflags).
:- use_module(b_enumerate).

% TO DO: call something which will use predicate_components

b_check_property_satisfiable(Properties,TypedConstants,ConstantsState,GlobalState) :- !,
     b_ast_cleanup:clean_up_pred(Properties,[],Pred),
     set_up_typed_localstate(TypedConstants,_FreshVars,TypedVals,[],ConstantsState,positive),
     b_interpreter_components:reset_component_info(false),
     append(ConstantsState,GlobalState,State),
     b_interpreter_components:b_trace_test_components(Pred,State,TypedVals),
     \+ b_interpreter_components:unsat_component_exists.
b_check_property_satisfiable(Properties,TypedConstants,ConstantsState,GlobalState) :-
    set_up_typed_localstate(TypedConstants,_FreshVars,TypedVals,[],ConstantsState,positive),
 %   print(checking_properties(TypedConstants,Properties,ConstantsState,GlobalState)),nl,
  %  print('CHECKING: '),
    init_wait_flags(WF),
    b_tighter_enumerate_values(TypedVals,WF),  % user tighter enumerate, rather than standard enumerate
    (GlobalState = []
      ->  b_test_boolean_expression(Properties,[],ConstantsState,WF)
      ;   b_test_boolean_expression(Properties,ConstantsState,GlobalState,WF)
    ),
    %ground_det_wait_flag(WF), print(grounding_wait_flags(ConstantsState,WF)),nl,
    ground_wait_flags(WF).



% ------------------------------------
% This code might be obsolete - replaced by unsat_cores module
% ------------------------------------

% NEW FIND UNSAT CORE

% Assumes predicate unsatisfiable without time-out to start with
% not yet finished

%:- use_module(bsyntaxtree, [conjunction_to_list/2,conjunct_predicates/2,find_identifier_uses/3]).

% :- use_module(predicate_debugger), predicate_debugger:unsat_core_goal.
%unsat_core_goal :-
%    bmachine:b_get_machine_goal(Goal),
%    unsat_core(Goal).
%unsat_core(FullPred) :-
%    get_parameters(FullPred,Params,Conj),
%    get_current_state('@PRED',CurID,CurBState),
%    set_error_context(unsat_core(CurID)),
%    conjunction_to_list(Conj,List),
%    create_texpr(truth,pred,[],T), append(List,[T],ListT), % so that we also check full property
%    unsat_core_list(ListT,[],Params,CurBState).

%remove_last(List,H,NewList) :- append(NewList,[H],List).

%unsat_core_list([],List,_Parameters,_CurBState) :-
%    conjunct_predicates(List,Pred),
%    print('UNSAT CORE: '), translate:print_bexpr(Pred),nl,
%    find_identifier_uses(Pred,[],Used),
%    print('USED IDENTIFIERS: '), print(Used),nl.
%unsat_core_list(List1,List2,Parameters,CurBState) :-
%    remove_last(List1,H,NewList1),
%    print('Checking removal of : '), translate:print_bexpr(H),nl,
%    append(NewList1,List2,List),
%    conjunct_predicates(List,Pred),
%    print(' --> '),translate:print_bexpr(Pred),nl,
%    preferences:get_preference(time_out,CurTO), DebugTimeOut is CurTO * 3,
%    tools:start_ms_timer(TIMER),
%    (time_out(b_check_property_satisfiable(Pred,Parameters,_ConstantsState,CurBState),
%             DebugTimeOut,TimeOutRes)
%     -> (TimeOutRes=time_out -> print('*** TIME OUT ***') ; print(' *** SAT ***')),nl,
%        tools:stop_ms_timer(TIMER),
%        unsat_core_list(NewList1,[H|List2],Parameters,CurBState)
%     ;  print(' --- STILL UNSAT (REMOVING CONJUNCT) ---'),nl,
%        tools:stop_ms_timer(TIMER),
%        unsat_core_list(NewList1,List2,Parameters,CurBState)
%    ).

