% (c) 2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(cbc_ba,[cbc_symbolic_invariant_violation/2,
   tcltk_cbc_symbolic_invariant_violation/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'A constraint-based invariant checker using before-after predicates.').

:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,
                                 b_get_properties_from_machine/1,
                                 b_specialized_invariant_for_op/2]).
:- use_module(probsrc(before_after_predicates), [before_after_predicate_for_operation/2]).
:- use_module(symbolic_model_checker(predicate_handling), [prime_predicate/2]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2,
                            create_negation/2]).
:- use_module(probsrc(preferences), [get_preference/2]).

:- use_module(probsrc(solver_interface), [solve_predicate/4]).
%:- use_module(symbolic_model_checker(solver_handling), [solve/2]).

cbc_symbolic_invariant_violation(OpName,Result) :-
    b_get_properties_from_machine(PropertiesOnConstants),
    b_get_invariant_from_machine(Inv), % TO DO?: split into conjuncts; use proof info
    prime_predicate(Inv,PrimedInv),
    create_negation(PrimedInv,NegInv),
    before_after_predicate_for_operation(OpName,BA),
    format('~nLooking for invariant violation for operation ~w~n',[OpName]),
    (get_preference(use_po,false) -> NegInvForOp=NegInv
     ; b_specialized_invariant_for_op(OpName,UncheckedInvariant)
     -> prime_predicate(UncheckedInvariant,PrimedUncheckedInv),
        create_negation(PrimedUncheckedInv,NegInvForOp)
      ; NegInvForOp=NegInv, print(no_specialized_invariant_info_for_op(OpName)),nl
    ),
    conjunct_predicates([PropertiesOnConstants,Inv,BA,NegInvForOp],Constraint),
    (debug:debug_mode(on) -> translate:nested_print_bexpr(Constraint),nl ; true),
    %solve(Constraint,Result).
    solve_predicate(Constraint,_State,1,Result).

% TO DO:
% static_symmetry_reduction_for_global_sets(State)
% (get_proven_invariant(OpName,ProvenInvariant) -> true

% use_module(symbolic_model_checker(cbc_ba)), cbc_symbolic_invariant_violation(Op,R).
% use_module(symbolic_model_checker(cbc_ba)), cbc_symbolic_invariant_violation(Op,R), R\=contradiction_found.
% use_module(symbolic_model_checker(cbc_ba)), findall((Op,R),cbc_symbolic_invariant_violation(Op,R),L), print(L),nl.

tk_call_cbc(O,R) :-
  cbc_symbolic_invariant_violation(O,Res), translate(Res,R).

translate(contradiction_found,R) :- !, R=ok.
translate(contradiction_found(_),R) :- !,R=ok.
translate(no_solution_found(_),R) :- !, R=ok_for_set_sizes.
translate(solution(_),R) :- !,R=invariant_violation.
translate(Res,F) :- functor(Res,F,_).


:- use_module(probsrc(tools),[start_ms_timer/1, stop_ms_timer/1]).
tcltk_cbc_symbolic_invariant_violation(list([list(['Operation','Status'])|Res])) :-
   start_ms_timer(Timer),
   findall(list([OpName,Result]),tk_call_cbc(OpName,Result),Res),
   stop_ms_timer(Timer).
