% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% disable clauses are not together warning, because
% cli_testcase/4 and cli_testcase/5 are sorted by id
:- set_prolog_flag(discontiguous_warnings,off).

%:- prolog_flag(compiling,_,debugcode).
%:- prolog_flag(source_info,_,on).
%:- prolog_flag(profiling,_,on).

/*
:- use_module('../extensions/profiler/profiler.pl').    % (adds the profiler himself)
:- use_module('../extensions/profiler/profiler_te.pl'). % (adds the term expander)
%:- use_module('../extensions/profiler/profiler_gui.pl'). % (adds the term expander)
:- enable_profiling_all(b_expression_sharing).
:- enable_profiling(bsyntaxtree:transform_bexpr_with_acc/5).
:- enable_profiling(bsyntaxtree:transform_bexpr/3).
:- enable_profiling(b_ast_cleanup:recompute_used_ids_info/2).
:- enable_profiling(bsyntaxtree:sub_expression_contains_wd_condition/1).
*/

:- ensure_loaded(prob_cli).

:- module(test_runner, [
                        test_repl/0,
                        run_last_test/0, last/0,
                        run_tests_by_id/1, id/1, ids/1,
                        run_tests_all/0, run_tests_all_except/1,
                        run_tests_by_category/1, category/1,
                        makefile_by_category/2, copy/1,
                        generate_makefile/2,
                        generate_makefile_all_tests/0,
                        full_flush/0,
                        jenkins_sanity_check/0,
                        sanity_check/1,
                        v/0, vv/0,
                        cache/0, % enable caching
                        all_categories/1,
                        all_ids/1,
                        ids_in_category/2
                       ]).

%:- use_module('../extensions/profiler/profiler.pl').

:- use_module(library(system)).
:- use_module(library(lists)).
:- use_module(library(sets), [intersect/2]).
:- use_module(library(file_systems)).

:- use_module(junit_tests,[print_junit/2,set_junit_dir/1,create_junit_result/6]).
:- use_module(testcases).
:- use_module(system_call).
:- use_module(tools,[safe_number_codes/2,get_options/5]).
:- use_module(tools,[host_platform/1,platform_is_64_bit/0]). % for conditions in testcases
:- use_module(tools_commands,[edit_file/1]).
:- use_module(error_manager,[get_total_number_of_errors/1, get_all_errors/1]).
:- use_module(performance_messages,[toggle_perfmessages/0]).

:- use_module(module_information).
:- module_info(group,testing).
:- module_info(description,'This module runs the tests stored in testcases.pl.').

:- use_module(library(sets)).

:- meta_predicate generate_makefile(-,1).
generate_makefile(Filename,WriteMakeTargetPredicate) :-
    tell(Filename),
    all_categories(AllTestCategories),
    write('testtarget: '),
    write_subtargets(AllTestCategories), nl, nl,
    maplist(WriteMakeTargetPredicate,AllTestCategories),
    told.

write_subtargets([G]) :- write(G).
write_subtargets([G|Gs]) :-
    write(G), write(' '),
    write_subtargets(Gs).

generate_makefile_all_tests :-
     generate_makefile('Makefile_All_Tests', write_makefile_target).
write_makefile_target(G) :- format('~w:\n\tsicstus -l src/test_runner.pl --goal "run_tests_by_category(\'~w\'), halt ; halt."',[G,G]),nl,nl.

all_categories(X) :-
    findall(Groups,
            cli_testcase(_Id,Groups,_Conf,_Args,_Desc),
            List),
    append(List,ListOfGroups),
    list_to_set(ListOfGroups,X).

all_ids(X) :-
    findall(Id,
            cli_testcase(Id,_Groups,_Conf,_Args,_Desc),
            List),
    list_to_set(List,X).

ids_in_category(Category,IDs) :-
    findall(Id,
            (cli_testcase(Id,Groups,_Conf,_Args,_Desc),member(Category,Groups)),
            List),
    list_to_set(List,IDs).

prob_junit(PROB_JUNIT) :-
    environ('PROB_JUNIT_DIR', Var), !,
    PROB_JUNIT = ['-junit', Var].
prob_junit([]).

dontstop(Var) :-
    environ('DONTSTOP', Var2), !, Var=Var2.
dontstop(false).

halt1_allowed :- \+ repl_mode,
   prob_junit([]), dontstop(false).

tests_multiply_timeout(Var) :-
    environ('TESTS_MULTIPLY_TIMEOUT', Var2), !, tools:arg_is_number(Var2,Var).
tests_multiply_timeout(Var) :- multiply_timeout(Var).

:- dynamic multiply_timeout/1.

% call updir from within Spider, when you are in the src subdirectory
set_cur_dir :- (directory_exists('./src') -> true /* we seem to be in the right directory */
          ; directory_exists('./cia') -> current_directory(_Old,'./..')
          ; print('Please start tests from ProB directory or ProB/src directory'),nl),
   print('Current directory: '),
   current_directory(CD), print(CD),nl.

reset_test_runner :- reset_nr_of_tests, set_cur_dir, retractall(test_failed(_)), retractall(test_diff_failed(_,_,_)).

ids(Nr) :- show_command(Nr).
show_command(Nr) :- cli_testcase(Nr,_,_,L,Desc),
  print('# '), print(Desc),nl, l_pr([probcli|L]).
l_pr([]) :- nl.
l_pr([H|T]) :- print(' '),print(H), l_pr(T).

:- volatile sanity_check_failed/0.
:- dynamic sanity_check_failed/0.
jenkins_sanity_check :-
   sanity_check(true),
   (sanity_check_failed -> halt(1) ; halt(0)).
sanity_check(_Jenkins) :-
   cli_testcase(Id,Categories,_,Args,_Comm1),
   \+ member('-strict',Args),
   print('*** test does not use -strict option: '), print_test(Id,Categories),nl,
   assert(sanity_check_failed), fail.
sanity_check(Jenkins) :-
   Jenkins == false,
   cli_testcase_diff_check_output(Id,_F1,_F2),
   cli_testcase(Id,Categories,_I1,_A1,_Comm1),
   (Categories = [_]
    -> true
    ; print('*** multiple categories for test with output file: '),
      print_test(Id,Categories),
      nl  /* can generate race conditions when run in parallel */),
   assert(sanity_check_failed), fail.
sanity_check(_Jenkins) :-
   cli_testcase_diff_check_output(Id,_F1,_F2),
   \+ (cli_testcase(Id,_Categories,_I1,_A1,_Comm1)),
   print('*** no cli_testcase, but diff output defined for id: '), print(Id),nl,
   assert(sanity_check_failed), fail.
sanity_check(_Jenkins) :-
   cli_testcase(Id,C1,I1,A1,Comm1),
   cli_testcase(Id,C2,I2,A2,Comm2),
   [C1,I1,A1,Comm1] \= [C2,I2,A2,Comm2],
   print('*** multiple cli_testcases for id: '), print(Id),nl,
   assert(sanity_check_failed), fail.
sanity_check(_Jenkins).

print_test(Id,Comment) :- print(Id),print(' : '), print(Comment).

% run the latest test (with highest id)
last :- run_last_test.
run_last_test :- get_last_test_id(Id),
  run_tests_by_id(Id).

get_last_test_id(Id) :- set_last_test_id, last_test_id(Id).

set_last_test_id :-
   retractall(last_test_id(_)), assert(last_test_id(-1)),
   cli_testcase(Id,C1,I1,A1,Comm1),
   ((cli_testcase(Id,C2,I2,A2,Comm2),(C1,I1,A1,Comm1) \= (C2,I2,A2,Comm2))
     -> nl,print('*** multiple entries for test id: '), print(Id),nl
     ; true),
   update_last(Id),
   fail.
set_last_test_id.

:- volatile last_test_id/1.
:- dynamic last_test_id/1.
update_last(Id) :- number(Id),!, % debug_print(9,Id), debug_print(9,' '),
   retract(last_test_id(L)),
   (Id>L -> Max = Id ; Max = L),
   assert(last_test_id(Max)).
update_last(Id) :- nl,print('*** test id not a number: '), print(Id),nl.

repeat_id(ID,Nr) :- repeat_id_aux(ID,Nr,10000000,0).
repeat_id_aux(ID,Nr,Min,Max) :- Nr>0, !, N1 is Nr-1,
  statistics(walltime,[Start,_]),
  id(ID),
  statistics(walltime,[Stop,_]),
  WT is Stop - Start,
  NewMin is min(Min,WT), NewMax is max(Max,WT),
  repeat_id_aux(ID,N1,NewMin,NewMax).
repeat_id_aux(_,_,Min,Max) :- format('Walltime Minumum = ~w ms, Maximum = ~w ms~n',[Min,Max]).

% run a test with a specific id
id(X) :- run_tests_by_id(X).
run_tests_by_id(X) :- reset_test_runner,
  run_tests_by_id_aux(X),
  check_failed.
run_tests_by_id_aux([]).
run_tests_by_id_aux(From-To) :- !,
    (From>To -> true
     ; run_tests_by_id_aux(From), F1 is From+1,
       run_tests_by_id_aux(F1-To)).
run_tests_by_id_aux([Id | Ids]) :- !,
    run_tests_by_id_aux(Id),
    run_tests_by_id_aux(Ids).
run_tests_by_id_aux(Id) :-
    cli_testcase(Id,_TestCategories,TestInfos,Arguments,Comment), !,
    run_list([testcase(Id,TestInfos,Arguments,Comment)]).
run_tests_by_id_aux(Id) :-
    \+(cli_testcase(Id,_Category,_Infos,_Arguments,_Comment)), !,
    format(user_output,'*** No testcase with ID ~w found~n', [Id]).

run_tests_all :- run_tests_all_except([]).

:- use_module(library(ordsets)).
run_tests_all_except(Categories) :- reset_test_runner,
    sort(Categories,SC),
    cli_testcase(Id,TestCategories,Infos,Arguments,Comment), sort(TestCategories,TC),
    \+ ord_intersect(TC,SC),
    run_list([testcase(Id,Infos,Arguments,Comment)]),
    fail.
run_tests_all_except(_) :- check_failed.


% run all tests with a specific category
category(X) :- run_tests_by_category(X).
run_tests_by_category(X) :- reset_test_runner,
    (is_list(X) -> run_tests_by_category_aux(X) ; run_tests_by_category_aux([X])),
    check_failed.
run_tests_by_category_aux(Categories) :-
    findall(testcase(Id, Infos, Arguments, Comment),
            (cli_testcase(Id, TestCategories, Infos, Arguments, Comment),
                intersect(Categories, TestCategories)),
            List),
    ( List=[]   -> format(user_output,'*** No testcase with categories ~w found~n', [Categories])
    ; otherwise -> run_list(List)).

run_tests_using_command(Command) :- reset_test_runner,
    findall(testcase(Id, Infos, Arguments, Comment),
            (cli_testcase(Id, _TestCategories, Infos, Arguments, Comment),
                member(Command, Arguments)),
            List),
    ( List=[]   -> format(user_output,'*** No testcase using command ~w found~n', [Command])
    ; otherwise -> run_list(List), check_failed).

run_tests_using_preference(Pref,Val) :- reset_test_runner,
    findall(testcase(Id, Infos, Arguments, Comment),
            (cli_testcase(Id, _TestCategories, Infos, Arguments, Comment),
                append(_,['-p',Pref,Val|_],Arguments) ),
            List),
    ( List=[]   -> format(user_output,'*** No testcase using preference ~w with value ~w found~n', [Pref,Val])
    ; otherwise -> run_list(List), check_failed).

% Generate file list (for copying) of a category :
show_files(Cat) :-
   cli_testcase(Id, TestCategories, _Infos, Arguments, _Comment),
   member(Cat, TestCategories),  % print(Arguments),nl,
   ( file_in_arguments(File,Arguments)
     ;
     cli_testcase_diff_check_output(Id,_GeneratedFile,File)),
   format('~w ',[File]),
   fail.
show_files(_) :- nl.

file_in_arguments(F,['-p',_,_|T]) :- !, file_in_arguments(F,T).
file_in_arguments(F,[C,_|T]) :- binary_command(C),!, file_in_arguments(F,T).
file_in_arguments(F,[H|T]) :- is_file(H),H=F ; file_in_arguments(F,T).
binary_command(eval). binary_command(evalt). binary_command(evalf).
binary_command('-mc').
is_file(F) :- atom(F), atom_codes(F,Codes), member(47,Codes),!.


% Generate a Makefile for all listed Categories
% example: makefile_by_category('AlstomMakefile',[animate,history,sptxt,rel_fnc])
% the Makefile can be run to perform the listed tests on a compiled version of probcli
makefile_by_category(File, Categories) :-
    (is_list(Categories) -> Cats = Categories ; Cats = [Categories]),
    my_open(File, Stream),
    format(Stream, 'PROBCLI=probcli~n',[]),
    format(Stream, 'test_all:', []),
    makefile_write_categories(Stream, Cats),
    makefile_by_category_aux(Stream, Cats),
    my_close(Stream).
makefile_by_category_aux(_Stream, []).
makefile_by_category_aux(Stream, [Cat | Cats]) :-
    makefile_by_category_single(Stream, Cat),
    makefile_by_category_aux(Stream, Cats).
makefile_by_category_single(Stream, Cat) :-
    findall(testcase(Id, Infos, Arguments, Comment),
            (cli_testcase(Id, TestCategories, Infos, Arguments, Comment),
                member(Cat, TestCategories)),
            List),
    (List=[] -> format(user_output,'*** No testcase with category ~w found~n', [Cat])
             ;  format(Stream, '~n~w:~n', [Cat]), makefile_write_calls(Stream, List)).

my_open(user_output,S) :- !, S=user_output.
my_open(File,S) :- open(File,write,S).
my_close(user_output) :- !.
my_close(S) :- close(S).

makefile_write_categories(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_categories(Stream, [Cat | Cats]) :-
    format(Stream, ' ~a', [Cat]),
    makefile_write_categories(Stream, Cats).

makefile_write_calls(_Stream, []).
makefile_write_calls(Stream, [testcase(Id, _Infos, Arguments, Comment) | Tests]) :-
    (Comment='' -> true ; format(Stream,'\techo \"Test ~w : ~w\"~n',[Id,Comment])),
    format(Stream, '\t$(PROBCLI)', []),
    makefile_write_arguments(Stream, Arguments),
    makefile_write_diff(Stream, Id),
    makefile_write_calls(Stream, Tests).

makefile_write_arguments(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_arguments(Stream, [Arg | Args]) :-
    format(Stream, ' ~w', [Arg]),
    (quote_next_arg(Arg) -> makefile_write_arguments_quoted(Stream,Args)
      ; makefile_write_arguments(Stream, Args)).

quote_next_arg('-goal'). % this will contain spaces ,... surround in "" for shell
quote_next_arg('--check_goal').
quote_next_arg('-check_goal').
quote_next_arg('-cbc_deadlock_pred').
quote_next_arg('-eval').
quote_next_arg('-evalt').
quote_next_arg('-evalf').
quote_next_arg('-cbc_sequence_with_target'). % actually quotes next two !
quote_next_arg('-cbc_sequence_with_target_all'). % ditto <- TO DO


makefile_write_arguments_quoted(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_arguments_quoted(Stream, [Arg | Args]) :-
    format(Stream, ' \"~w\"', [Arg]),
    makefile_write_arguments(Stream, Args).

makefile_write_diff(Stream, ID) :-
    cli_testcase_diff_check_output(ID, File1, File2),
    format(Stream, '\tdiff -b ~w ~w~n', [File1, File2]),
    fail.
makefile_write_diff(_Stream, _ID).

% -------------------------

copy(Cat) :- (Cat=[_|_] -> C=Cat ; C=[Cat]),
   generate_copy_commands(C,'testarchive/').

:- use_module(b_trace_checking,[get_default_trace_file/2]).
generate_copy_commands(Categories,Dest) :-
    cli_testcase(ID, TestCategories, _Infos, Arguments, _Comment),
    non_empty_inter(Categories,TestCategories), %print(ID),nl,
    Arguments=[MainFile|_], generate_copy_command(MainFile,Dest), % print(MainFile),nl,
    additional_testcase_file(ID,MainFile,Arguments,ExtraFile),
    generate_copy_command(ExtraFile,Dest),
    fail.
generate_copy_commands(_,_).

additional_testcase_file(ID,_,_,EFile) :- extra_testcase_file(ID,EFile).
additional_testcase_file(_ID,File,Arguments,TFile) :- member('-t',Arguments),
    get_default_trace_file(File,TFile).
additional_testcase_file(ID,_,_,RefFile2) :- cli_testcase_diff_check_output(ID,_File1,RefFile2).

non_empty_inter(A,B) :- member(X,A), member(X,B),!.

:- use_module(tools_strings,[string_concatenate/3]).
:- use_module(tools,[get_parent_directory/2]).
generate_copy_command(File,Dest) :-
    safe_file_exists(File),
    get_parent_directory(File,Dir),!,
    string_concatenate(Dest,Dir,DestDir),
    string_concatenate(Dest,File,DestFile),
    format(user_output,'\tmkdir -p ~w~n',[DestDir]),
    get_command_path(mkdir,MkCmdPath),
    system_call(MkCmdPath, ['-p',DestDir],_Text1,_JExit1),
    format(user_output,'\tcp ~w ~w~n',[File,DestFile]),
    get_command_path(cp,CpCmdPath),
    system_call(CpCmdPath, [File,DestFile],_Text2,_JExit2).
generate_copy_command(_,_).

% -------------------------

:- dynamic test_failed/1, test_diff_failed/3.

check_failed :- test_failed(X),!,print_failed_tests,
   print('Use the following command to run individual tests: '),nl,
   print('   sicstus -l src/test_runner.pl --goal "id('),print(X),print(')."'),nl,
   (halt1_allowed -> halt(1) ; print('halt(1) :: junit / dontstop mode'),nl).
check_failed :- number_of_tests_run(Nr),
   (Nr=1 -> format('Test successful.~n',[])
    ; format('All ~w tests successful.~n',[Nr])).
print_failed_tests :- number_of_tests_run(Nr),
   findall(Y,test_failed(Y),Fails), length(Fails,NrFails),
   format('** Tests run: ~w, failed: ~w **~n** Failed tests:~n',[Nr,NrFails]),
   test_failed(X), print(X), print(' '), lookup_test_description(X,Desc), print(Desc), nl, fail.
print_failed_tests :- nl.

lookup_test_description(Id,Desc) :- cli_testcase(Id,_,_,_,Desc).

run_list(List) :- maplist(run_single_testcase,List).


cache :- add_additional_arguments(['-cache','/Users/leuschel/svn_root/NewProB/examples/cache/']).
v :- add_additional_arguments(['-v']). % verbose
vv :- add_additional_arguments(['-vv']). % very_verbose

:- volatile additional_arguments/1.
:- dynamic additional_arguments/1.
% add additional cli arguments when running tests:
add_additional_arguments(List) :-
  (retract(additional_arguments(Old)) -> true ; Old=[]),
   append(Old,List,New),
   assert(additional_arguments(New)).

% add additional preference when running tests:
add_additional_preference(PREF,PREFVAL) :-
  (retract(additional_arguments(Old0)) -> remove_matching_pref(Old0,PREF,Old) ; Old=[]),
   New = ['-p',PREF,PREFVAL|Old],
   format('New additional arguments: ~w~n',[New]),
   assert(additional_arguments(New)).

% remove all preferences conflicting with PREF
remove_matching_pref([],_PREF,[]).
remove_matching_pref([H|T],PREF,[H|TT]) :- non_pref_cli_arg(H),!,
   remove_matching_pref(T,PREF,TT).
remove_matching_pref(['-p',P,OLD|T],PREF,Res) :- !,
   (P=PREF -> Res=T ; Res = ['-p',P,OLD|RT], remove_matching_pref(T,PREF,RT)).
remove_matching_pref([H|T],PREF,[H|RT]) :- remove_matching_pref(T,PREF,RT).

non_pref_cli_arg('-v').
non_pref_cli_arg('-vv').

% remove all preferencs conflicting with other list of prefs
remove_matching_prefs([],P,P).
remove_matching_prefs([H|T],InPrefs,Res) :- non_pref_cli_arg(H),!,
   remove_matching_prefs(T,InPrefs,Res).
remove_matching_prefs(['-cache',_File|T],InPrefs,Res) :- !,
   remove_matching_prefs(T,InPrefs,Res).
remove_matching_prefs(['-p',PREF,_|T],InPrefs,Res) :-
   remove_matching_pref(InPrefs,PREF,In2),
   remove_matching_prefs(T,In2,Res).

skip(Infos) :-
    member(skip,Infos) -> true
    ; member(conditional_skip(Callable), Infos), call(Callable).

% RUNNING SINGLE TESTCASE:
% ------------------------
run_single_testcase(testcase(Id,Infos,Arguments,_Comment)) :-
    skip(Infos), !,
    full_flush,
    format(user_output,'Skipping test ~w ~n', [Id]), print_junit_skip(Arguments),
    full_flush.
run_single_testcase(testcase(Id,_Infos,Arguments,Comment)) :-
    full_flush,
    format(user_output,'Running test ~w ~n      ~w~n',[Id,Comment]),
    print('testcase: probcli '), print_args(Arguments),nl,
    clear_diff_output(Id),
    prob_junit(JUnit),
    append(Arguments, JUnit, Arguments1),
    (additional_arguments(ExtraArgs)
     -> remove_matching_prefs(ExtraArgs,Arguments1,RemArguments1),
        append(ExtraArgs,RemArguments1,Arguments2)
     ; Arguments1=Arguments2),
    (tests_multiply_timeout(Factor) -> modify_timeout(Factor,Arguments2,Arguments3) ; Arguments3 = Arguments2),
    print('executed: probcli '), print_args(Arguments3),nl,
    full_flush,
    test_started(Id),
    catch(user:go_cli(Arguments3), Exception, true), !,
    test_finished(Id),
    (
      Exception == halt(0)                        -> check_diff_output(Id) ;
      var(Exception)                              -> check_diff_output(Id) ;
      Exception = error(X,Y)                      -> (halt1_allowed -> format(user_output,'error on test execution: error(~w,~w)',[X,Y]), halt(1) ; test_failed(Id,Exception)) ;
      Exception = enumeration_warning(_,_,_,_,_)  -> (halt1_allowed -> format(user_output,'exception on test execution: ~w)',[Exception]), halt(1) ; test_failed(Id,Exception)) ;
      Exception = solver_and_provers_too_weak     -> (halt1_allowed -> format(user_output,'exception on test execution: ~w)',[Exception]), halt(1) ; test_failed(Id,Exception)) ;
      Exception == halt(1)                        -> test_failed(Id, Exception) ;
      Exception == parse_errors(_PE)              -> test_failed(Id, Exception) ;
      otherwise                                   -> test_failed(Id, Exception), format(user_output,'Exception not caught in test_runner: ~w~n',[Exception]), halt(1)
     ), !,
    user:reset_cli,
    full_flush.

:- dynamic test_took_aux/1.
:- dynamic last_testcase_run/1, number_of_tests_run/1.
number_of_tests_run(0).
reset_nr_of_tests :- retractall(number_of_tests_run(_)), assert(number_of_tests_run(0)).

:- use_module(library(system),[now/1, datime/2]).
:- dynamic performance_session_running/1, performance_session_stats/4.

performance_session_start :-
    now(When),
    datime(When,datime(Year,Month,Day,Hour,Min,Sec)),
    format('~nStarting Codespeed Performance Monitoring session ~w:~w:~w:~w:~w:~w~n',[Year,Month,Day,Hour,Min,Sec]),
    retractall(performance_session_running(_)),
    assert(performance_session_running(When)).

:- use_module(parsercall,[get_parser_version/1]).
:- use_module(version, [version_str/1, revision/1, lastchangeddate/1]).
performance_session_end(FilePrefix) :-
    performance_session_running(When),
    datime(When,datime(Year,Month,Day,Hour,Min,Sec)),
    %tools:ajoin([FilePrefix,':',Year,Month,Day,Hour,Min,Sec],FileName),
    format('~nFinishing Codespeed session ~w:~w:~w:~w:~w:~w~n -> File : ~w~n',[Year,Month,Day,Hour,Min,Sec,FilePrefix]),
    open(FilePrefix,append,S),
    format(S,'~n/* Codespeed session ~w:~w:~w:~w:~w:~w */~n',[Year,Month,Day,Hour,Min,Sec]),
    version_str(Vers), portray_clause(S, session_prob_version(When,Vers)),
    revision(Rev), portray_clause(S, session_prob_revision(When,Rev)),
    lastchangeddate(DD), portray_clause(S, session_prob_lastchangeddate(When,DD)),
    get_parser_version(PV), portray_clause(S, session_prob_parser_version(When,PV)),
    write_perf_data(When,S).

write_perf_data(When,S) :- additional_arguments(New),
    portray_clause(S, stored_additional_arguments(When,New)),
    fail.
write_perf_data(When,S) :- performance_session_stats(When,Id,Time,WTime),
    portray_clause(S, stored_performance_test_stats(When,Id,Time,WTime)),
    fail.
write_perf_data(_When,S) :- nl(S), nl(S), close(S).


test_started(Id) :-
    retractall(last_testcase_run(_)), assert(last_testcase_run(Id)),
    retractall(test_took_aux(_)),
    statistics(runtime,[Start,_]),
    statistics(walltime,[WStart,_]),
    bb_put(test_started,Start),
    bb_put(test_started_wtime,WStart).

:- dynamic test_stats/5.
test_finished(Id) :-
    statistics(runtime,[End,_]),
    statistics(walltime,[WEnd,_]),
    bb_get(test_started,Start),
    bb_get(test_started_wtime,WStart),
    Time is End - Start, WTime is WEnd- WStart,
    (retract(number_of_tests_run(Nr)) -> N1 is Nr+1 ; N1=1),
     assert(number_of_tests_run(N1)),
    (retract(test_stats(Id,PrevTime,PrevWTime,_,_))
       -> assert(test_stats(Id,Time,WTime,PrevTime,PrevWTime))
        ; assert(test_stats(Id,Time,WTime,-1,-1))
    ),
    (performance_session_running(When)
     -> assert(performance_session_stats(When,Id,Time,WTime))
     ;  true).

print_delta_stats :- print('Comparing walltimes with previous test run: '),nl,
   findall(delta(DeltaPerc,DeltaWTime,Id),test_delta_stat(Id,DeltaPerc,DeltaWTime),L),
   (L=[] -> print('No previous run information available'),nl
    ; print(' ID |  % (delta absolute) | walltime (runtime)~n'),nl,
      sort(L,SL),
      maplist(print_delta,SL)).
test_delta_stat(Id,DeltaPerc,DeltaWTime) :-
   test_stats(Id,_RTime,WTime,_PrevRTime,PrevWTime),
   PrevWTime>0,
   DeltaWTime is WTime - PrevWTime,
   DeltaPerc is (100*DeltaWTime) / PrevWTime.
print_delta(delta(DeltaPerc,DeltaWTime,Id)) :-
   test_stats(Id,RTime,WTime,_PrevRTime,PrevWTime),
   format(' ~w | ~2f % (~w ms) | ~w ms (~w ms runtime) [~w walltime ms previously]~n',
          [Id,DeltaPerc,DeltaWTime,WTime,RTime,PrevWTime]).


test_took(X) :- test_took_aux(X), !.
test_took(XInSeconds) :-
    statistics(runtime,[End,_]),
    bb_get(test_started,Start),
    X is End - Start,
    XInSeconds is X / 1000,
    asserta(test_took_aux(XInSeconds)).

print_junit_skip(Arguments) :-
    prob_junit(['-junit', Dir])
    -> set_junit_dir(Dir), test_took(X),
       create_junit_result(Arguments,'Integration Tests','Integration_Test',X,skip,Result),
       print_junit([Result],'Integration_Test')
    ;  true.

% if the test expects a time_out error, the timeout is not expanded
% otherwise, timeout is increased to allow coverage analysis / junit / etc to finish
modify_timeout(_,OldOptions,OldOptions) :- segment(OldOptions,['-expcterr','time_out']), !.
modify_timeout(Factor,[],['-p','TIME_OUT',NVal]) :- % timeout was not set at all - set it to Factor*Default
   % Note: there is a potential problem when the time_out is set inside the machine and not in the test !! TO DO: fix
   preferences:preference_default_value(time_out,DEFAULT),
   NVal is Factor * DEFAULT.
modify_timeout(Factor,['-p','TIME_OUT',OLD|T],['-p','TIME_OUT',NewT|T]) :-
   tools:arg_is_number(OLD,OLDT), !,
   % preferences:preference_default_value(time_out,DEFAULT),
   % OLDT < DEFAULT,!, % we explicitly set a TIME_OUT lower than the default value
   NewT is OLDT * Factor.
modify_timeout(Factor,[H|T],[H|MT]) :- modify_timeout(Factor,T,MT).

full_flush :- flush_output(user_output), flush_output(user_error).

print_args([]).
print_args([H|T]) :- print(H), print(' '), print_args(T).

test_failed(Id,Msg) :-
  test_failed(Id,Msg,'').
test_failed(Id,Msg1,Msg2) :-
  test_took(X),
  cli_testcase(Id,Categories,_Infos,Arguments,Comment),
  Categories = [FirstCat|_], !,
  ErrorMessage = ['Test with Id',Id,'failed.','Test Categories: ',Categories,'Test Arguments: ',Arguments,'Test Comment: ',Comment,'Error Messages in Test Runner: ',Msg1,Msg2,'Content of Error Manager: '],
  (get_all_errors(AllErrors) -> true ; AllErrors = []),
  append(ErrorMessage,AllErrors,FullErrorMessage),
  create_junit_result(Id,'Integration Tests',FirstCat,X,error(FullErrorMessage),Result),
  print_junit([Result],'Integration_Test'),
  format(user_output,'*** Test ~w FAILED: ~w~w~n', [Id,Msg1,Msg2]),
  assert(test_failed(Id)).


diff_failed(Id,F1,F2,DiffOutputText,CmpOutputText) :-
  test_took(X),
  atom_codes(DiffOutputAtom,DiffOutputText),
  atom_codes(CmpOutputAtom,CmpOutputText),
  ErrMsg = ['Diff for test with Id\n',Id,'\nfailed:\n','Output file\n',F1,'\ndoes not correspond to stored version\n',F2,
            '\nOutput of Diff:\n',DiffOutputAtom,
            '\nOutput of Cmp:\n',CmpOutputAtom],
  create_junit_result(Id,'Diff Checking','Diff_checking',X,
              error(ErrMsg),
              Result),
  print_junit([Result],'Diff_checking'),
  format(user_output,'*** Test ~w FAILED: Diff failed:~nOutput file ~w~ndoes not correspond to stored version~n~w~n', [Id,F1,F2]),
  format(user_output,'Diff:~n~s~n',[DiffOutputText]),
  format(user_output,'Cmp:~n~s~n',[CmpOutputText]),
  assert(test_failed(Id)),
  assert(test_diff_failed(Id,F1,F2)).

:- use_module(tools_commands,[diff_files_with_editor/2]).
diff_in_editor :- findall(I,test_diff_failed(I,_,_),LI), sort(LI,SI),
  format('Opening failed diff files in editor: ~w~n',[SI]),
  test_diff_failed(Id,F1,F2),
  format('Test ~w~n~w ~w~n',[Id,F1,F2]),
  diff_files_with_editor(F1,F2),
  fail.
diff_in_editor.


clear_diff_output(Id) :- % clear all files that should be generated
  cli_testcase_diff_check_output(Id,GeneratedFile,_StoredReferenceFile),
  safe_file_exists(GeneratedFile),
  (cli_testcase_do_not_delete(Id,GeneratedFile) -> format(user_output,'% Keeping: ~w~n',[GeneratedFile])
   ; format(user_output,'% Deleting: ~w~n',[GeneratedFile]),
     delete_file(GeneratedFile)
  ),fail.
clear_diff_output(_).

check_diff_output(Id) :-
  findall(diff(Id,GeneratedFile,StoredReferenceFile),
          cli_testcase_diff_check_output(Id,GeneratedFile,StoredReferenceFile),
      ListOfDiffsToCheck),
  maplist(check_diff_output2, ListOfDiffsToCheck).

check_diff_output2(diff(Id,GeneratedFile,StoredReferenceFile)) :-
  \+ safe_file_exists(GeneratedFile)       -> test_failed(Id,'Output file does not exist:',GeneratedFile) ;
  \+ safe_file_exists(StoredReferenceFile) -> test_failed(Id,'Stored file does not exist:',StoredReferenceFile) ;
  otherwise                           -> diff(Id,GeneratedFile,StoredReferenceFile).

diff(Id,F1,F2) :-
  format(user_output,'% Checking: diff / cmp ~w ~w~n',[F1,F2]),
  get_command_path(diff,DiffPath),
  get_command_path(cmp,CmpPath),
  (system_call(DiffPath,['-b',F1,F2],DiffOutputText,_ErrTextDiff,ExitDiff) % use -q for quiet
    -> true
     ; DiffOutputText = "*** CALLING DIFF FAILED !", ExitDiff = fail
  ),
  format(user_output,'% Checking: cmp ~w ~w~n',[F1,F2]),
  (system_call(CmpPath,['-b',F1,F2],CmpOutputText,_ErrTextCmp,_ExitCmp)
    -> true
     ; CmpOutputText = "*** CALLING CMP FAILED !"
  ),
  (ExitDiff = exit(0)%, ExitCmp = exit(0)
   -> true
   ; diff_failed(Id,F1,F2,DiffOutputText,CmpOutputText)).

:- volatile repl_mode/0.
:- dynamic repl_mode/0.

:- use_module(library(lists),[maplist/2]).
% a minimal shell to execute tests:
test_repl :- assert(repl_mode),
   prolog_flag(argv,ArgV), maplist(test_eval_argv,ArgV),
   test_repl_loop,
   retractall(repl_mode).
test_eval_argv(Cmd) :- format('ARGV ==> ~w~n',[Cmd]),
   atom_codes(Cmd,C), safe_number_codes(Nr,C), !, test_eval(Nr).
test_eval_argv(Cmd) :- test_eval(Cmd),!.

test_repl_loop :- print('TEST ==> '),read(T), test_eval(T), !, test_repl_loop.
test_repl_loop.

:- meta_predicate wall(0).
wall(Call) :-
    statistics(walltime,[Start,_]),
    call(Call),
    statistics(walltime,[Stop,_]), WT is Stop-Start,
    format('Walltime: ~w ms~n',[WT]).

% -------------------------

:- use_module(library(file_systems)).
test_file(Id,File,AbsFileName) :- cli_testcase(Id,_Cat,_Infos,Arguments,_Comment),
   get_options(Arguments,user:recognised_cli_option,_Options,Files,fail),
   member(File,Files),
   is_existing_file(File),
   absolute_file_name(File,AbsFileName).

is_existing_file(X) :- \+ number(X), atom(X),
   atom_codes(X,Codes),[BS] = "/", (member(BS,Codes) -> true),
   file_exists(X).

all_files(Files) :- findall(F,test_file(_,_,F),A), sort(A,Files).

traverse :- traverse('../prob_examples/public_examples/').
traverse(SD) :- all_files(Files), absolute_file_name(SD,StartDir),
    format('Examining files in ~w~n +  means file is used in some test~n~n',[StartDir]),
    traverse(StartDir,Files).

traverse(Dir,AllFiles) :- file_member_of_directory(Dir,_,FullFile),
    tools:get_filename_extension(FullFile,XT),
    (member(FullFile,AllFiles) -> format(' +  ~w~n',[FullFile])
      ; relevant_extension(XT) -> format('--- ~w~n',[FullFile])),
    fail.
traverse(Dir,AllFiles) :- directory_member_of_directory(Dir,_,SubDir),
    %format('~nSTART ~w~n',[SubDir]),
    traverse(SubDir,AllFiles),
    %format('~n  END ~w~n',[SubDir]),
    fail.
traverse(_,_).

relevant_extension('mch').
relevant_extension('ref').
relevant_extension('imp').
relevant_extension('tla').
relevant_extension('fuzz').
relevant_extension('csp').
relevant_extension('cspm').
relevant_extension('eventb').
% --------------------------

test_eval(N) :- number(N),!, wall(id(N)).
test_eval(last) :- !, wall(run_last_test).
test_eval(N-M) :- number(N), number(M),!, wall(id(N-M)).
test_eval(repeat(ID,M)) :- !, repeat_id(ID,M).
test_eval(v) :- !,v.
test_eval(verbose) :- !,v.
test_eval(all_files) :- !, all_files(Files), format('~nFiles = ~n~w~n',[Files]).
test_eval(files) :- !, traverse.
test_eval(files(Dir)) :- !, traverse(Dir).
test_eval(cache) :- !,print('Enabling cache'),nl,
    cache.
test_eval(debug) :- !,print('Enabling Prolog debugging mode (use -v or -vv for ProB debugging info)'),nl,
    debug,
    retractall(multiply_timeout(_)),
    assert(multiply_timeout(10)).
test_eval(timeout(X)) :- !,
    retractall(multiply_timeout(_)),
    assert(multiply_timeout(X)).
test_eval(debug_off) :- !,print('Disabling Prolog debugging mode'),nl,
    nodebug,
    retractall(multiply_timeout(_)).
test_eval(clpfd) :- !,print('Enabling CLPFD'),nl,
    add_additional_preference('CLPFD','TRUE').
test_eval(clpfd_off) :- !,print('Disabling CLPFD'),nl,
    add_additional_preference('CLPFD','FALSE').
test_eval(smt) :- !,print('Enabling SMT'),nl,
    add_additional_preference('SMT','TRUE').
test_eval(smt_off) :- !,print('Disabling SMT'),nl,
    add_additional_preference('SMT','FALSE').
test_eval(chr) :- !,print('Enabling CHR'),nl,
    add_additional_preference('CHR','TRUE').
test_eval(chr_off) :- !,print('Disabling CHR'),nl,
    add_additional_preference('CHR','FALSE').
test_eval(cse_off) :- !,print('Disabling CSE'),nl,
    add_additional_preference('CSE','FALSE').
test_eval(cse) :- !,print('Enabling CSE'),nl,
    add_additional_preference('CSE','TRUE').
test_eval(cse_subst) :- !,print('Enabling CSE_SUBST'),nl,
    add_additional_preference('CSE','TRUE'),
    add_additional_preference('CSE_SUBST','TRUE').
test_eval(trace_info) :- !,print('Enabling TRACE_INFO'),nl,
    add_additional_preference('TRACE_INFO','TRUE').
test_eval(p(PREF)) :- !,print('Enabling Preference '),print(PREF),nl,
    add_additional_preference(PREF,'TRUE').
test_eval(p(PREF,VAL)) :- !,print('Setting Preference '),print(PREF),nl,
    add_additional_preference(PREF,VAL).
test_eval(random) :- !,print('Enabling RANDOMISE_ENUMERATION_ORDER'),nl,
    add_additional_preference('RANDOMISE_ENUMERATION_ORDER','TRUE').
test_eval(random_off) :- !,print('Enabling RANDOMISE_ENUMERATION_ORDER'),nl,
    add_additional_preference('RANDOMISE_ENUMERATION_ORDER','FALSE').
test_eval(sanity_check) :- !, sanity_check(false).
test_eval(sc) :- !, sanity_check(false).
test_eval(trace) :- !, print('Enabling TRACE_UPON_ERROR'),nl,
    add_additional_preference('TRACE_UPON_ERROR','TRUE').
test_eval(trace_off) :- !, print('Disabling TRACE_UPON_ERROR'),nl,
    add_additional_preference('TRACE_UPON_ERROR','FALSE').
test_eval(raise) :- !,print('Enabling STRICT_RAISE_ENUM_WARNINGS'),nl,
    add_additional_preference('STRICT_RAISE_ENUM_WARNINGS','TRUE').
test_eval(nopt) :- !,print('Disabling OPTIMIZE_AST'),nl,
    add_additional_preference('OPTIMIZE_AST','FALSE').
test_eval(vv) :- !,vv.
test_eval(q) :- !,fail.
test_eval(x) :- !,halt.
test_eval(reload) :- !,use_module(probsrc(test_runner)), use_module(probsrc(testcases)).
test_eval(edit) :- last_testcase_run(Id),
   cli_testcase(Id,_,_Infos,Arguments,_Comment),
   member(File,Arguments),  safe_file_exists(File),!,
   edit_file(File).
test_eval(e) :- !, test_eval(edit).
test_eval(diff) :- !, diff_in_editor.
test_eval(quit) :- !,fail.
test_eval(halt) :- !,fail.
test_eval(info) :- !, get_total_number_of_errors(X), format('~nTotal number of errors: ~w~n~n',[X]).
test_eval(end_of_file) :- !,fail.  % Ctrl-D
test_eval(cat) :- !, print('Categories: '),
   findall(Cat,
           (cli_testcase(_, TestCategories, _, _, _),  member(Cat, TestCategories)), List),
   sort(List,SL), print(SL),nl.
test_eval(cat(Category)) :- !,
   wall(run_tests_by_category(Category)).
test_eval(make(Categories)) :- !,
   wall(makefile_by_category(user_output,Categories)).
test_eval(make(File,Categories)) :- !,
   wall(makefile_by_category(File,Categories)).
test_eval(files(Category)) :- !, show_files(Category).
test_eval(uses(Command)) :- !, wall(run_tests_using_command(Command)).
test_eval(uses(Pref,Val)) :- !, wall(run_tests_using_preference(Pref,Val)).
test_eval(p) :- !, test_eval(profile).
test_eval(ps) :- !, test_eval(profile_stats).
test_eval(delta) :- !, print_delta_stats.
test_eval(start) :- !, performance_session_start.
test_eval(stop) :- !, performance_session_end('log/test_runner_performance_log.pl').
test_eval(codespeed) :- !,
   performance_session_start,
   test_eval(cat(codespeed)),
   performance_session_end('log/test_runner_performance_codespeed_log.pl').
test_eval(profile) :- !, print('PROFILING : '), %spy([avl:avl_size/2]),
   (current_prolog_flag(profiling,on)
    -> set_prolog_flag(profiling,off), print('OFF') ;
       set_prolog_flag(profiling,on), print('ON')),
    nl,print('USE ps to print profile info'),nl.
test_eval(profile_stats) :- !, nl,print('PROFILE INFORMATION:'), nl,
   on_exception(error(existence_error(_,_),_),print_profile, print('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl, debug:timer_statistics.
test_eval(perf) :- !, toggle_perfmessages.
test_eval(Category) :- valid_category(Category),!,
   wall(run_tests_by_category(Category)).
test_eval([Category|T]) :- valid_category(Category),!,
   wall(run_tests_by_category([Category|T])).
test_eval(C) :-
     (C=help -> true ; print('*** Unknown command'), nl),
     print(' Commands:   Nr,  Nr-Nr,  last, q, x, v, vv, uses(Cmd), uses(Pref,Val),'),nl,
     print('             repeat(id,nr), timeout(factor), e,edit,diff,'),nl,
     print('             cat, cat(Cats),make(Cats),make(File,Cats),files(Cat),'),nl,
     print('             profile,cache,debug, trace,trace_off, perf, reload, sc,'),nl,
     print('             clpfd,smt,chr,cse,cse_subst,random,clpfd_off, nopt, delta, info, ...'),nl.

safe_file_exists(F) :- atom(F), file_exists(F).

valid_category(Cat) :-
   cli_testcase(_Id, TestCategories, _Infos, _Arguments, _Comment),
   member(Cat, TestCategories).
